/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 * 
 * High speed IP packet redundancy elimination using rabin fingerprinting.
 * Although based on the description of Spring & Wetherall, the devil is in
 * the details --- e.g. data structures for recovery, multi-core pipelining,
 * recovery from loss and packet re-ordering.
 */

#include "util.h"
#include "wqueue.h"
#include "rabinpoly.h"
#include "mersene.h"

//#define BIND_WORKQUEUES_TO_CPUS
//#define BIND_WORKQUEUES_TO_CORES
#define EXHAUSTIVE_DIFF_PREFIX
//#define EXHAUSTIVE_DIFF_PREFIX_PREFETCH /* useless */
#define EXHAUSTIVE_DIFF_POSTFIX
#define USE_KMEM_CACHE_WORK /* use a special purpose cache */
#define RESEQUENCE_IP_HDRS /* re-sequence all IP traffc in the encoder */

#define DDUP_LOG 						"DEDUP"
#define SUPERSCALAR_FACTOR				4
#define SKB_ALLOC_MASK					GFP_ATOMIC /* DMA ready patch SKBs */
//#define NF_HOOK_OUT_SKB
//#define EXPLICIT_IP_ROUTE_HARD
//#define MARSHALL_DIFF /* saves 2 bytes per patch header */
//#define IP_FORWARD_ONLY_THROUGH_WQUEUES /* simple forward only */
//#define FAKE_DIFF /* for memory access accnting, do not do diffs */

/* the ~ values are valid for TCP & UDP, all we care about */
#define IPPROTO_DEDUP(proto)			((~(proto)) & 0xff)
#define IPPROTO_IPDEDUP_UDP				IPPROTO_DEDUP(IPPROTO_UDP) 
#define IPPROTO_IPDEDUP_TCP				IPPROTO_DEDUP(IPPROTO_TCP)
#define ONE_K 							(1UL << 10)
#define ONE_MEG							(1UL << 20)
#define ETH_MTU 						1500
#define ETH_JUMBO_MTU					9000
#define DEFAULT_MTU						ETH_MTU
/* Ashok & Akella showed smaller window (42) is necessary for popular matches */
#define DEFAULT_RABIN_WINDOW			64

/* 6 yields on average 21.6 fingerprints, 7 - 10.8, 8 - 5.4, and 9 yields 2.7,
 * if 0, all (MTU+1-DEFAULT_RABIN_WINDOW) fingerprints will be considered */
#define DEFAULT_RABIN_REPR_BITS			8//7
#define SEC_OF_1GBPS_DATA				(125 * ONE_MEG)
#define DEFAULT_RF_IDX_BUCKETS 			4

/* 2^19 is fine, 2^16 is about 94MB for ETH_MTU */
#define DEFAULT_PACKET_STORE_COUNT		((6*(SEC_OF_1GBPS_DATA))/DEFAULT_MTU)

#define estimate_avg_fp(size, window, bits) \
				( ((size) + 1 - (window)) / (1UL << (bits)) )
#define AVG_FP estimate_avg_fp(DEFAULT_MTU, DEFAULT_RABIN_WINDOW, \
				DEFAULT_RABIN_REPR_BITS)
#define FIXED_FP						16//16
#define DEFAULT_RF_INDEX_COUNT 			(FIXED_FP * DEFAULT_PACKET_STORE_COUNT)
#define COUNTERS_RESET_STRING			"reset-counters"
#define MERSENE_TWISTER_SEED			0xdeadbeef

int logical_cpu_map[] = 
	{0, 8, 4, 12, 1, 9, 5, 13, 2, 10, 6, 14, 3, 11, 7, 15};
static int cores = SUPERSCALAR_FACTOR;
module_param(cores, int, 0000);
MODULE_PARM_DESC(cores, "The superscalar factor (number of CPUs/cores)");
static int shard_udp_port = 6969;
static int rabin_window = DEFAULT_RABIN_WINDOW;
module_param(rabin_window, int, 0000);
MODULE_PARM_DESC(rabin_window, "The number of bytes a rabin fingerprint is "
		"computed over");
static int rabin_representative_bits = DEFAULT_RABIN_REPR_BITS;
module_param(rabin_representative_bits, int, 0000);
MODULE_PARM_DESC(rabin_representative_bits, "The number of least significant "
		"bits equal to zero for a rabin fingerprint to be representative");
static int mtu = DEFAULT_MTU;
module_param(mtu, int, 0000);
MODULE_PARM_DESC(mtu, "The MTU of the MAC layer");
static long rf_index_buckets = DEFAULT_RF_IDX_BUCKETS;
module_param(rf_index_buckets, long, 0000);
MODULE_PARM_DESC(rf_index_buckets, "Index buckets for dedup");
static long packet_store_count = DEFAULT_PACKET_STORE_COUNT;
module_param(packet_store_count, long, 0000);
MODULE_PARM_DESC(packet_store_count, "Packet store count for dedup");
static long rf_index_count = DEFAULT_RF_INDEX_COUNT;
module_param(rf_index_count, long, 0000);
MODULE_PARM_DESC(rf_index_count, "Index count for dedup");

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 24)
static int hook_chain = NF_IP_FORWARD;
#else
static int hook_chain = NF_INET_FORWARD;
#endif
module_param(hook_chain, int, 0000);
MODULE_PARM_DESC(hook_chain, "Netfilter hook chain");
static int hook_prio = NF_IP_PRI_FIRST;
module_param(hook_prio, int, 0000);
MODULE_PARM_DESC(hook_prio, "Netfilter hook priority");
static int filter_port = 0; /* 0 means ALL traffic is FECed */
module_param(filter_port, int, 0000);
MODULE_PARM_DESC(filter_port, "The port on which traffic is intercepted on, "
		"0 means all but UDP FEC port");
static char* no_iface = "";
module_param(no_iface, charp, 0000);
MODULE_PARM_DESC(no_iface, "Interface that should not be interfered with");
static char* pipe_iface = "";
module_param(pipe_iface, charp, 0000);
MODULE_PARM_DESC(pipe_iface, "Interface on which incoming packets on the "
		"filter port should be buffered for FEC");
static char *procfile = "ipdedup";
module_param(procfile, charp, 0000);
MODULE_PARM_DESC(procfile, "The name of the /proc/<file>");
static char *mirror_ip = NULL;
module_param(mirror_ip, charp, 0000);
MODULE_PARM_DESC(mirror_ip, "The IP address of the mirroring host");
static char *local_ip = NULL;
module_param(local_ip, charp, 0000);
MODULE_PARM_DESC(local_ip, "The IP address of the mirroring host");

struct dedup_pipe
{
	int pipe_iface, pipe_iface_ifindex, no_iface, no_iface_ifindex;
	struct sockaddr_in addr;
	struct sockaddr_in remote;
};

/* reports for the entire stream of IP packets */
struct ddup_reports
{
	unsigned int ipfrags;
	unsigned int noiface;
	unsigned int at_encoder;
	unsigned int at_decoder;
	unsigned int nomatch;
	unsigned int skb_nonlinear;
};

struct diff
{
	u_int64_t fingerprint;
	u_int16_t offset, bytes_before, bytes_after;
};

/* marshalled diff, saves 2 bytes for each (struct diff) */
struct diff_m
{
	u_int16_t fingerprint[4], offset, bytes_before, bytes_after;
};

static inline void diff_marshall(struct diff *diff, struct diff_m *mdiff)
{
	mdiff->fingerprint[0] = diff->fingerprint >> (3 * 16);
	mdiff->fingerprint[1] = (diff->fingerprint >> (2 * 16)) & 0xffff;
	mdiff->fingerprint[2] = (diff->fingerprint >> (16)) & 0xffff;
	mdiff->fingerprint[3] = (diff->fingerprint) & 0xffff;

	mdiff->offset = diff->offset;
	mdiff->bytes_before = diff->bytes_before;
	mdiff->bytes_after = diff->bytes_after;
}

static inline void diff_unmarshall(struct diff *diff, struct diff_m *mdiff)
{
	diff->fingerprint = ( ((u_int64_t) mdiff->fingerprint[0]) << (3*16) ) |
		( ((u_int64_t) mdiff->fingerprint[1]) << (2*16) ) | 
		( ((u_int64_t) mdiff->fingerprint[2]) << (16) ) |
		( (u_int64_t) mdiff->fingerprint[3] );
	
	diff->offset = mdiff->offset;
	diff->bytes_before = mdiff->bytes_before;
	diff->bytes_after = mdiff->bytes_after;
}

struct rf
{
	u_int64_t fingerprint;
	u_int16_t offset, len;
};

struct rf_index_entry
{
	struct rf f;
	u_int32_t pkt_id;
};

struct rf_index
{
	u_int32_t cnt;
	unsigned char buckets;
	struct rf_index_entry *entries;
	u_int32_t updates, collision, overwrite_invalid, same_fingerprint;
};

struct packet_store
{
	u_int32_t cnt;
	unsigned short pkt_size;
	char *pkt_store; /* format: | size (of type u_int16_t) | size bytes | */
};

struct fingerprint_pad
{
	struct rf *pad;
	struct diff *diff;
	int pad_size, diff_size, repr_bits;
	u_int64_t repr_bitmask;
	u_int32_t fp, repr_fp; /* count the number of fingerprints */
	struct mersene_twister rand;
};

struct rf_state
{
	u_int32_t next_pkt_id;
	struct packet_store pkts;
	struct rf_index rfidx;
	struct rabinpoly_window rw;
	struct fingerprint_pad fp;
	int64_t bytes_saved;
	u_int64_t total_bytes;
	/* the difference between the two is lost packets (inconsistent caches) */
	unsigned int patches, total_diffs, recovered_diffs, encoded_diffs, 
		unrecoverable_diffs, corrupted_diffs;
};

struct proc_dir_entry *proc_entry = NULL;
struct nf_hook_ops dedup_nf_hook_ops;
struct task_struct **workers = NULL;
struct dedup_pipe pipe_state;
struct ddup_reports ddup_rep;
void *procfile_page = NULL; /* a page to read / write the procfile */
struct fifo_wqueue **wqueue = NULL;
struct rf_state *rf_st = NULL;

#ifdef USE_KMEM_CACHE_WORK
struct kmem_cache *dedup_skb_work_cache = NULL;
#ifndef KMEM_CACHE
#define KMEM_CACHE(__struct, __flags) kmem_cache_create(#__struct,\
				sizeof(struct __struct), __alignof__(struct __struct),\
				(__flags), NULL)
#endif /* KMEM_CACHE */
#endif

/*
 * Don't actually neeed it, and I shaved off some 10% workqueue usage.
 * Remember, ip_route_me_harder expects skb->dst to be set.
 */
static inline int do_route_me_harder(struct sk_buff **pskb)
{
#ifdef EXPLICIT_IP_ROUTE_HARD
	unsigned addr_type = RTN_UNSPEC;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
	return (ip_route_me_harder(pskb, addr_type));
#else
	return (ip_route_me_harder((*pskb), addr_type));
#endif /* LINUX_VERSION_CODE */
#else
	return 0;
#endif /* EXPLICIT_IP_ROUTE_HARD */
}

static void destroy_rf_state(struct rf_state *st)
{
	struct packet_store *ps = &st->pkts;
	struct rf_index *ri = &st->rfidx;
	
	/* free the packet store */
	if (ps->pkt_store)
	{
		vfree(ps->pkt_store);
		ps->pkt_store = NULL;
	}
	
	/* free the index */
	if (ri->entries)
	{
		vfree(ri->entries);
		ri->entries = NULL;
	}
	
	if (st->fp.pad)
		kfree(st->fp.pad);
	if (st->fp.diff)
		kfree(st->fp.diff);
	memset(&st->fp, 0, sizeof(struct fingerprint_pad));
	
	rabinpoly_deinit(&st->rw);
}

#define MAX_UINT_32	0xffffffff /* this is probably defined somewhere else */

static int init_rf_state(struct rf_state *st, u_int32_t packet_store_count,
		unsigned short packet_size, u_int32_t rf_index_count, 
		unsigned short rf_index_buckets, int rf_window_size, int fp_repr_bits)
{
	int err = 0, i;
	size_t bytes;
	struct packet_store *ps;
	struct rf_index *ri;
	
	memset(st, 0, sizeof(struct rf_state));
	
	/* first the packet store */
	ps = &st->pkts; 
	ps->cnt = packet_store_count;
	ps->pkt_size = packet_size;
	
	bytes =  ps->cnt * (sizeof(u_int16_t) + (ps->pkt_size * sizeof(char)));
	if (!(ps->pkt_store = vmalloc(bytes)))
	{
		printk(DDUP_LOG": vmalloc failure (->pkt_store) {%u bytes}\n", 
				(unsigned) bytes);
		err = -ENOMEM;
		goto out;
	}
	/* zero it out, touching the pages to fix the mapping (in page fault) */
	memset(ps->pkt_store, 0, bytes);
	
	/* next the index */
	ri = &st->rfidx;
	ri->cnt = rf_index_count;
	ri->buckets = rf_index_buckets;
	
	bytes = ri->cnt * ri->buckets * sizeof(struct rf_index_entry);
	if (!(ri->entries = vmalloc(bytes)))
	{
		printk(DDUP_LOG": vmalloc failure (->entries) {%u bytes}\n", 
				(unsigned) bytes); 
		err = -ENOMEM;
		goto out;
	}
	/* initialize the rf_index_entries, important to figure out expired */
	memset(ri->entries, 0, bytes);
	for (i = 0; i < ri->cnt * ri->buckets; i++)
		ri->entries[i].pkt_id = (((u_int32_t) (0-(1+ps->cnt))) % MAX_UINT_32);
	
	/* the fingerprint pad */
	bytes = (ps->pkt_size + 1 - rf_window_size) * sizeof(struct rf);
	if (!(st->fp.pad = kmalloc(bytes, GFP_KERNEL)))
	{
		printk(DDUP_LOG": alloc falure (->fp.pad {%u bytes})\n", 
				(unsigned) bytes);
		err = -ENOMEM;
		goto out;
	}
	/* the fingerprint diff pad */
	bytes = (ps->pkt_size + 1 - rf_window_size) * sizeof(struct diff);
	if (!(st->fp.diff = kmalloc(bytes, GFP_KERNEL)))
	{
		printk(DDUP_LOG": alloc falure (->fp.diff {%u bytes})\n", 
				(unsigned) bytes);
		err = -ENOMEM;
		goto out;
	}
	if (rabin_representative_bits == 0)
	{
		st->fp.repr_bits = 0;
		st->fp.repr_bitmask = 0x0;
	}
	else
	{
		st->fp.repr_bits = rabin_representative_bits;
		st->fp.repr_bitmask = (u_int64_t)((1UL << rabin_representative_bits)-1);
	}
	st->fp.pad_size = 0;
	st->fp.diff_size = 0;
	
	pseudorand_init(&st->fp.rand, (int)MERSENE_TWISTER_SEED);
	if ((err = rabinpoly_init(&st->rw, FINGERPRINT_PT, rf_window_size)) != 0)
	{
		printk(DDUP_LOG": Error encountered: rabinpoly init\n");
		goto out;
	}
	
	out: //
	if (err != 0)
		destroy_rf_state(st);
	return err;
}

struct dedup_skb_work
{
	struct work_struct_lite work;
	struct sk_buff *skb;
	int at_decoder;
	struct rf_state *rfs;
};

/* returns a packet store location at position @idx, no checks */
static inline char *pkt_stored_at_idx(struct packet_store *ps, u_int32_t idx)
{
	int offset = idx * ps->pkt_size;
	return &(ps->pkt_store[offset]); /* this is a byte offset */
}

/* returns the buckets (all ->buckets) at hashed position @idx, no checks */
static inline struct rf_index_entry *get_rf_idx(struct rf_index *ri, 
		u_int32_t idx)
{
	int offset = idx * ri->buckets;
	return &(ri->entries[offset]); /* offsets every sizeof(rf_index_entry) */
}

/* much less expensive than keeping an inverted index */
static inline int is_index_entry_valid(struct rf_state *rfs, u_int32_t p)
{
	u_int32_t m = (rfs->next_pkt_id - 1), T = rfs->pkts.cnt;
	if (m >= T)
		return ( (p <= m) && (p >= (m-T)) );
	else
		return ( (p <= m) || (p >= ((m-T) % MAX_UINT_32)) );
}

/* does not fail, overwrites the current slot */
static inline void add_skb_to_store(struct rf_state *rfs, struct sk_buff *skb, 
		int offset, int len)
{
	struct packet_store *ps = &rfs->pkts;
	/* I wonder how expensive are the modulo division operations */
	char *slot = pkt_stored_at_idx(ps, rfs->next_pkt_id % ps->cnt);
	/* add the packet size --- stored in the first 2 bytes */
	u_int16_t *slot_size = (u_int16_t *) slot;
	*slot_size = len;
	slot += sizeof(u_int16_t); /* pointer arithmetic stuff */
	/* we assume linear skb, however skb_copy_bits works either way */
	skb_copy_from_linear_data_offset(skb, offset, slot, len);
	/* NOT % size, otherwise can't tell (fast) if an index entry is valid */
	rfs->next_pkt_id++;
}

/* seek over the ip and tcp/udp headers, no checks */
static inline unsigned char *tcp_udp_payload(struct iphdr *iph)
{
	if (iph->protocol == IPPROTO_UDP || iph->protocol == IPPROTO_IPDEDUP_UDP)
	{
		struct udphdr *uh = ip_udp_hdr(iph);
		return (char *) ((void *) uh + sizeof(struct udphdr));
	}
	else if (iph->protocol == IPPROTO_TCP || 
			iph->protocol == IPPROTO_IPDEDUP_TCP)
	{
		struct tcphdr *th = ip_tcp_hdr(iph);
		return (char *) ((void *) th + (th->doff * 4));
	}
	else
		BUG_ON(1); /* this should never be reached */
}

static inline int is_characteristic_fp(struct fingerprint_pad *p, u_int64_t fp)
{
	/* worst case scenario, all fingerprints are considered */
	return !(fp & p->repr_bitmask);
	/* using the mersene twister random generator I get > 30% drop in tput */
	//return !(genrand_int32(&p->rand) & p->repr_bitmask);
}

/* returns the number of fingerprints */
static int fingerprint_packet(struct rf_state *rfs, 
		u_char *data, int len)
{
	struct rabinpoly_window *rw = &rfs->rw;
	struct fingerprint_pad *fp = &rfs->fp;
	int i, rabin_window = get_rabinpoly_window_size(rw);
	
	fp->pad_size = 0; /* this has to be set before returning */
	if (unlikely(len <= 0))
		goto out;

	/* start pristine */
	rabinpoly_reset(rw);
	/* first block or all bytes if less */
	for (i = 0; i < min(len, rabin_window); i++)
		rabinpoly_slide8(rw, data[i]);
	
	if (is_characteristic_fp(fp, rw->fingerprint))
	{
		fp->pad[fp->pad_size].offset = 0;
		fp->pad[fp->pad_size].len = min(len, rabin_window);
		fp->pad[fp->pad_size].fingerprint = rw->fingerprint;
		fp->pad_size++;
		fp->repr_fp++;
	}
	fp->fp++;
	
	for (i = rabin_window; i < len; i++)
	{
		rabinpoly_slide8(rw, data[i]);
		/* have the compiler arrange the code for the common case */
		if (is_characteristic_fp(fp, rw->fingerprint))
		{
			fp->pad[fp->pad_size].offset = i + 1 - rabin_window;
			fp->pad[fp->pad_size].len = rabin_window;
			fp->pad[fp->pad_size].fingerprint = rw->fingerprint;
			fp->pad_size++;
			fp->repr_fp++;
		}
		fp->fp++;
	}
	out:
	return fp->pad_size;
}

/* the fingerprints are themselves random numbers, use them as hash values */
static inline u_int32_t hash_rf(struct fingerprint_pad *p, u_int64_t fp, 
			u_int32_t tablesize)
{
#if BITS_PER_LONG == 32
	u_int32_t high = fp >> 32;
	u_int32_t low = fp & MAX_UINT_32;
	return (high ^ low) % tablesize;
#elif BITS_PER_LONG == 64
	/* get rid of the least significant 0 bits */
	return (fp >>= p->repr_bits) % tablesize;
#else
#error Wordsize not 32 or 64
#endif
}

static void add_fp_to_index(struct rf_state *rfs)
{
	struct fingerprint_pad *fp = &rfs->fp;
	struct rf_index *rfidx = &rfs->rfidx;
	struct rf_index_entry *entries, *entry;
	int i, packet_id = (rfs->next_pkt_id - 1);
	
	for (i = 0; i < fp->pad_size; i++)
	{
		int j;
		struct rf *f = &(fp->pad[i]);
		entries = get_rf_idx(rfidx, hash_rf(fp, f->fingerprint, rfidx->cnt));
		
		for (j = 0; j < rfidx->buckets; j++)
		{
			entry = &entries[j];
			if (entry->f.fingerprint == f->fingerprint)
			{
				rfidx->same_fingerprint++;
				goto match_found;
			}
			if (!is_index_entry_valid(rfs, entry->pkt_id))
			{
				rfidx->overwrite_invalid++;
				goto match_found;
			}
		}
		entry = &entries[0]; /* otherwise pick the victim, in FIFO order */
		rfidx->collision++; /* doesn't account for warm-up */
		
		match_found: /* entry is set at this point */
		entry->f.fingerprint = f->fingerprint;
		entry->pkt_id = packet_id;
		entry->f.offset = f->offset;
		entry->f.len = f->len;
		rfidx->updates++;
	}
}

/* compute diff by expanding from the anchor both ``left and right'' */
static void do_diff(struct rf_state *rfs, struct diff *patch, 
		struct rf_index_entry *ref, char *payload, int len)
{
	struct packet_store *ps = &rfs->pkts;
	char *ref_bytes = pkt_stored_at_idx(ps, ref->pkt_id % ps->cnt);
	u_int16_t ref_size = *((u_int16_t *) ref_bytes);
	char *p, *p_bound, *r, *r_bound;
	
	ref_bytes += sizeof(u_int16_t);
	
	/* prefix bytes */
	patch->bytes_before = 0;
#ifdef EXHAUSTIVE_DIFF_PREFIX
	p = &payload[patch->offset - 1]; /* first byte before */
	p_bound = payload;
	r = &ref_bytes[ref->f.offset - 1];
	r_bound = ref_bytes;

#ifdef EXHAUSTIVE_DIFF_PREFIX_PREFETCH /* turns out it's useless */
	prefetch_range(p_bound, patch->offset);
	prefetch_range(r_bound, ref->f.offset);
#endif /* EXHAUSTIVE_DIFF_PREFIX_PREFETCH */
	
	while ( (p >= p_bound) && (r >= r_bound) && (*p == *r) )
	{
		patch->bytes_before++;
		p--;
		r--;
	}
#endif /* EXHAUSTIVE_DIFF_PREFIX */
	
	/* postfix bytes --- careful, the length of the block is not necessarily 
	 * the same, e.g. if a packet is < than standard rf block size */
	patch->bytes_after = ref->f.len;
	
#ifdef EXHAUSTIVE_DIFF_POSTFIX
	p = &payload[patch->offset + patch->bytes_after];
	p_bound = &payload[len];
	r = &ref_bytes[ref->f.offset + patch->bytes_after];
	r_bound = &ref_bytes[ref_size];
	
	while( (p < p_bound) && (r < r_bound) && (*p == *r) )
	{
		patch->bytes_after++;
		p++;
		r++;
	}
#endif /* EXHAUSTIVE_DIFF_POSTFIX */
}

/* returns true (1) if (x0, x1) overlaps (y0, y1), open intervals */
static inline int open_intervals_overlap(unsigned long x0, unsigned long x1, 
		unsigned long y0, unsigned long y1)
{
	return !((y1 <= x0) || (y0 >= x1));
}

static int entry_overlaps_existing_patch(struct diff *patch, 
		struct rf *f)
{
	return open_intervals_overlap(
			patch->offset - patch->bytes_before, 
			patch->offset + patch->bytes_after, 
			f->offset, f->offset + f->len);
}

int patch_overlaps_existing_patch(struct diff *patch, struct diff *new)
{
	return open_intervals_overlap(
			patch->offset - patch->bytes_before, 
			patch->offset + patch->bytes_after, 
			new->offset - new->bytes_before, 
			new->offset + new->bytes_after);
}

static int patch_overlap_existing_patch_fixup(struct diff *old_patch, 
		struct diff *new_patch)
{
	/* The offset of the new patch is beyond the old patch tail */
	long overlap = (old_patch->offset + old_patch->bytes_after) - 
			(new_patch->offset - new_patch->bytes_before);
	
	if (overlap > 0)
	{
		new_patch->bytes_before -= overlap;
		return 1;
	}
	return 0;
}

#define diff_patch_len(patch)	((patch->bytes_before) + (patch->bytes_after))

struct diff_hdr
{
	u_int16_t cnt; /* the number of (struct diff) following */
	u_int16_t len; /* the length in bytes of the original packet */
};

static void pkt_diffs(struct rf_state *rfs, struct sk_buff **pskb, 
		u_char *payload, int offset, int len)
{
	struct fingerprint_pad *fp = &rfs->fp;
	struct rf_index *rfidx = &rfs->rfidx;
	struct rf_index_entry *entries, *ref;
	struct sk_buff *nskb = NULL;
	struct diff *prev_patch = NULL;
	int i, skb_pos = 0;

	fp->diff_size = 0; /* start accumulating diffs */
	for (i = 0; i < fp->pad_size; i++)
	{
		int j;
		struct rf *f = &(fp->pad[i]);
		struct diff *patch = &fp->diff[fp->diff_size];
		
		entries = get_rf_idx(rfidx, hash_rf(fp, f->fingerprint, rfidx->cnt));
		for (j = 0; j < rfidx->buckets; j++)
		{
			ref = &entries[j];
			if ((ref->f.fingerprint == f->fingerprint) &&
					(is_index_entry_valid(rfs, ref->pkt_id)))
				goto match_found;
		}
		continue; /* to the next fingerprint */
		match_found: /* entry is set at this point */
		
		/* Matching while considering the overlap was quite annoying to get
		 * right, in the end we have a relatively efficient and fast method */
		if (!prev_patch)
		{
			/* the new skb must have plenty of room even for tiny payload */
			int skb_sz = LL_MAX_HEADER + max_t(int, skb_headlen(*pskb),
					sizeof(char) + sizeof(u_int16_t) + sizeof(struct diff) +
					get_rabinpoly_window_size(&rfs->rw));
			if (unlikely(!(nskb = alloc_skb(skb_sz, SKB_ALLOC_MASK))))
			{
				printk(DDUP_LOG": Unable to allocate diff skb\n");
				goto out_fail;
			}
			skb_reserve(nskb, LL_MAX_HEADER);
			skb_reset_network_header(nskb);
			nskb->dst = dst_clone((*pskb)->dst);
			
			/* prepare new skb, copy the header (ip, udp/tcp) up to payload */
			skb_put(nskb, offset);
			skb_copy_to_linear_data(nskb, (*pskb)->data, offset);
			skb_pos = offset;
			
			patch->offset = f->offset; /* do_diff needs this set */
			do_diff(rfs, patch, ref, payload, len);
		}
		else
		{
			/* f must not overlap the previous patch */
			if (entry_overlaps_existing_patch(prev_patch, f))
				continue;
			
			patch->offset = f->offset; /* do_diff needs this set */
			do_diff(rfs, patch, ref, payload, len);

			/* the new patch must not overlap the previous patch, fix it */
			patch_overlap_existing_patch_fixup(prev_patch, patch);
		}
		patch->fingerprint = f->fingerprint; /* one more memory access */
		fp->diff_size++;
		prev_patch = patch;
		rfs->patches++;
		/* preemptively wrap up, fingerprints are in increasing order */
		if (unlikely(patch->offset + patch->bytes_after > len))
		{
			printk(DDUP_LOG": Invalid patch range: %d + %d > %d\n", 
					patch->offset, patch->bytes_after, len);
			goto out_fail;
		}
		if ((patch->offset + patch->bytes_after) == len)
			break;
	}
	
	if (nskb)
	{
		struct iphdr *iph = skb_ip_hdr(nskb);
		int bytes_diffed, bytes_left = len, sz;
		struct diff_hdr dh;
		unsigned char *ref_bytes = payload;
		
		dh.cnt = fp->diff_size;
		dh.len = len;

		/* put the patch header (i.e. the metadata) */
		skb_put(nskb, sizeof(struct diff_hdr));
		skb_copy_to_linear_data_offset(nskb, skb_pos, &dh, sizeof(dh));
		skb_pos += sizeof(struct diff_hdr);
		
		/* put the patch data, the serialized versions */
#ifdef MARSHALL_DIFF
		sz = sizeof(struct diff_m);
		for (i = 0; i < fp->diff_size; i++)
		{
			struct diff_m diff_m;
			struct diff *diff = &(fp->diff[i]);
			
			diff_marshall(diff, &diff_m);
			skb_put(nskb, sz);
			skb_copy_to_linear_data_offset(nskb, skb_pos, &diff_m, sz);
			skb_pos += sz;
		}
#else
		sz = fp->diff_size * sizeof(struct diff);
		skb_put(nskb, sz);
		skb_copy_to_linear_data_offset(nskb, skb_pos, fp->diff, sz);
		skb_pos += sz;
#endif /* MARSHALL_DIFF */
		
		/* put the skb data w/o the patches */
		bytes_diffed = 0;
		for (i = 0; i < fp->diff_size; i++)
		{
			struct diff *patch = &fp->diff[i];
			int bytes = (patch->offset - patch->bytes_before) - bytes_diffed;
			
			skb_put(nskb, bytes);
			skb_copy_to_linear_data_offset(nskb, skb_pos, ref_bytes, bytes);
			skb_pos += bytes;
			
			bytes_diffed += (bytes + diff_patch_len(patch));
			ref_bytes = &payload[bytes_diffed];
		}
		/* don't forget about the tail, if any */
		BUG_ON(len < bytes_diffed);
		bytes_left = len - bytes_diffed;
		skb_put(nskb, bytes_left);
		skb_copy_to_linear_data_offset(nskb, skb_pos, ref_bytes, bytes_left);
		skb_pos += bytes_left;
		
		/* measure the savings */
		rfs->bytes_saved += ((int)ntohs(iph->tot_len) - (int)skb_pos);
		
		/* change the protocol apropriately, and adjust the length */
		iph->protocol = IPPROTO_DEDUP(iph->protocol);
		iph->tot_len = htons(skb_pos);
		
		/* checksum the packet */
#ifndef RESEQUENCE_IP_HDRS
		iph->check = 0;
		iph->check = ip_fast_csum(skb_network_header(nskb), iph->ihl);
		nskb->ip_summed = CHECKSUM_NONE; /* computed the checksum ourselves */
#endif /* RESEQUENCE_IP_HDRS */
		
		rfs->encoded_diffs++;
		rfs->total_diffs++;
		
		/* nf_ct_attach(nskb, *pskb); */
		kfree_skb(*pskb);
		*pskb = nskb;
		return;
		
		out_fail:
		if (nskb)
			kfree_skb(nskb);
	}
}

/* return NULL if the index points to an invalid packet store entry */
static char *get_reference_pkt(struct rf_state *rfs, struct diff *patch, 
		u_int16_t *ref_offset)
{
	int i;
	struct rf_index *rfidx = &rfs->rfidx;
	struct packet_store *ps = &rfs->pkts;
	struct fingerprint_pad *fp = &rfs->fp;
	struct rf_index_entry *entries, *entry;
	char *ref_bytes;
	u_int16_t ref_size;
	
	entries = get_rf_idx(rfidx, hash_rf(fp, patch->fingerprint, rfidx->cnt));
	for (i = 0; i < rfidx->buckets; i++)
	{
		entry = &entries[i];
		if ((entry->f.fingerprint == patch->fingerprint) && 
				(is_index_entry_valid(rfs, entry->pkt_id)))
			goto match_found;
	}
	printk(DDUP_LOG": Match not found for fingerprint %llu\n", 
			patch->fingerprint);
	return NULL;
	match_found: //
	
	*ref_offset = entry->f.offset;
	ref_bytes = pkt_stored_at_idx(ps, entry->pkt_id % ps->cnt);
	ref_size = *((u_int16_t *) ref_bytes);
	ref_bytes += sizeof(u_int16_t);
	
	/* double check the validity of the patch */
	if (unlikely((entry->f.offset - patch->bytes_before < 0) || 
			(entry->f.offset + patch->bytes_after > ref_size)))
	{
		printk(DDUP_LOG": Finngerprint reference range invalid: "
				"%d: %d, %d, %d (%d)\n", entry->f.offset, patch->offset, 
				patch->bytes_before, patch->bytes_after, ref_size);
		return NULL;
	}
	return ref_bytes;
}

/* offset is basically the size of the ip + the size of the udp/tcp headers */
static int pkt_patch(struct rf_state *rfs, struct sk_buff **pskb, 
		u_char *payload, int offset, int len)
{
	/* compute the size of the skb */
	struct sk_buff *nskb = NULL;
	struct iphdr *iph;
	struct diff_hdr *dh = (struct diff_hdr *) payload;

#ifdef MARSHALL_DIFF
	struct diff_m *diffs = (struct diff_m *)(payload + sizeof(struct diff_hdr));
#else
	struct diff *diffs = (struct diff *)(payload + sizeof(struct diff_hdr));
#endif /* MARSHALL_DIFF */

	int bytes_remaining = dh->len, i, skb_pos, patched_bytes;
	
	/* create a sufficiently large skb */
	if (unlikely(!(nskb = alloc_skb(LL_MAX_HEADER + offset + dh->len, 
			SKB_ALLOC_MASK))))
	{
		printk(DDUP_LOG": Unable to allocate diff skb\n");
		rfs->unrecoverable_diffs++;
		goto out_fail;
	}
	
	skb_reserve(nskb, LL_MAX_HEADER);
	skb_reset_network_header(nskb);
	nskb->dst = dst_clone((*pskb)->dst);
	
	/* copy the iphdr and udp or tcp header */
	skb_put(nskb, offset);
	skb_copy_to_linear_data(nskb, (*pskb)->data, offset);
	skb_pos = offset;
	
	/* effectively patch the packet payload */
#ifdef MARSHALL_DIFF
	payload += (sizeof(struct diff_hdr) + (dh->cnt * sizeof(struct diff_m)));
#else
	payload += (sizeof(struct diff_hdr) + (dh->cnt * sizeof(struct diff)));
#endif /* MARSHALL_DIFF */

	patched_bytes = 0; /* bytes patched so far */
	for (i = 0; i < dh->cnt; i++)
	{
		u_int16_t ref_off;
		int bytes, gap_bytes;
		char *ref_bytes;

#ifdef MARSHALL_DIFF
		struct diff_m *patch_m = &diffs[i];
		struct diff automatic_var_patch, *patch = &automatic_var_patch;
		diff_unmarshall(patch, patch_m);
#else
		struct diff *patch = &diffs[i];
#endif /* MARSHALL_DIFF */

		bytes = (patch->offset - patch->bytes_before) - patched_bytes;
		gap_bytes = diff_patch_len(patch);
		/* get_reference_pkt returns the offset of the fingerprint */
		ref_bytes = get_reference_pkt(rfs, patch, &ref_off);
		
		if (unlikely(!ref_bytes))
		{
			printk(DDUP_LOG": Cannot find valid recovery reference packet\n");
			rfs->unrecoverable_diffs++;
			goto out_fail;
		}
		
		/* bytes from the packet payload */
		skb_put(nskb, bytes);
		skb_copy_to_linear_data_offset(nskb, skb_pos, payload, bytes);
		skb_pos += bytes;
		payload += bytes;
		
		/* bytes from the reference packet */
		skb_put(nskb, gap_bytes);
		skb_copy_to_linear_data_offset(nskb, skb_pos, 
				&ref_bytes[ref_off - patch->bytes_before], gap_bytes);
		skb_pos += gap_bytes;
		
		/* adjust counts & offsets */
		bytes_remaining -= (bytes+gap_bytes);		
		patched_bytes += (bytes+gap_bytes);
	}
	/* don't forget about the tail */	
	skb_put(nskb, bytes_remaining);
	skb_copy_to_linear_data_offset(nskb, skb_pos, payload, bytes_remaining);
	skb_pos += bytes_remaining;
	
	/* revert back to the protocol, and the size */
	iph = skb_ip_hdr(nskb);
	/* compute the savings at the decoder site */
	rfs->bytes_saved += ((int)skb_pos - (int)ntohs(iph->tot_len));
	iph->protocol = IPPROTO_DEDUP(iph->protocol);
	iph->tot_len = htons(skb_pos);
	
	/* the cool thing is that with the TCP checksum I can figure out if my
	 * recovery messed up or not */
	if (iph->protocol == IPPROTO_TCP)
	{
		//struct tcphdr *th = ip_tcp_hdr(iph);
		int iphdr_len = iph->ihl * 4;
		struct tcphdr *th = (struct tcphdr *)((void *)iph + iphdr_len);
		int tcp_pack_len = (offset + dh->len) - iphdr_len;
		__sum16 check = th->check; /* save the old checksum */
		
		th->check = 0;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
		th->check = tcp_v4_check(th, tcp_pack_len, iph->saddr, iph->daddr,
				csum_partial((char *) th, tcp_pack_len, 0));
#else
		th->check = tcp_v4_check(tcp_pack_len, iph->saddr, iph->daddr,
				csum_partial((char *)th, tcp_pack_len, 0));
#endif
		if (unlikely(check != th->check))
		{
			printk(DDUP_LOG": !Recovery failed, mangled packet\n");
			rfs->corrupted_diffs++;
			goto out_fail;
		}
	}
	
	/* checksum the IP header, since the protocol and length changed */
	iph->check = 0;
	iph->check = ip_fast_csum(skb_network_header(nskb), iph->ihl);
	nskb->ip_summed = CHECKSUM_NONE; /* computed the checksum ourselves */		

	/* nf_ct_attach(nskb, *pskb); */
	kfree_skb(*pskb);
	*pskb = nskb;
	return 1;
	
	out_fail:
	if (nskb)
		kfree_skb(nskb);
	return 0;
}

#ifdef RESEQUENCE_IP_HDRS
/* If NAPI is used, then we are safe to increment this value, otherwise have
 * to provide an atomic_t like implementation for the short datatype. */
static unsigned short packet_id_cnt = 0;
/* wraps around when overflowing */
static inline unsigned short seq_inc_return(void) { return (packet_id_cnt++); }
#endif /* RESEQUENCE_IP_HDRS */

/*
 * Collisions in the hashspace are practically none for 64bit hashes and
 * the range of memory sizes we used (DEFAULT_RABIN_WINDOW). In the highly
 * unlikely event of a collision having occured, payload checksum 
 * (e.g. TCP checksum) or end-2-end bit error detection should kick in.
 * 
 * 
 * Here's how this works. Upon receiving packet p:
 * - if at encoder:
 * 		- put p in packet store
 * 		- compute representative fingerprints rfs of p
 * 		- for each rf in rfs:
 * 			- if rf in index:
 * 				- compute diff, merge non-overlapping diffs at this stage
 * 			- store rf in index, if bucket collision (== or full), replace 
 * 		- if diffs:
 * 			- prepare diff packet (IPPROTO_IPDEDUP, new size, re compute csum)
 * 	- else, if at decoder:
 * 		- if p is a diff packet:
 * 			- recover original packet from diff packet
 * 			- if recovery not succesful:
 * 				- initiate ``window'' cleanup
 * 		- goto (at encoder)
 * 	- forward packet p (this may be the diff packet if at encoder, or patched
 * 			packet if at decoder, or simply the original packet).
 */
static void dedup_skb_work_fn(struct work_struct_lite *work)
{
	struct dedup_skb_work *wk = container_of(work, struct dedup_skb_work, work);
	struct sk_buff *skb = wk->skb; /* the sk_buff has been NF_STOLEN */
	int at_decoder = wk->at_decoder;
	struct rf_state *rfs = wk->rfs;
	struct iphdr *iph = skb_ip_hdr(skb);
	unsigned char *pkt_payload = tcp_udp_payload(iph);
	int offset = pkt_payload - skb->data;
	int len = skb->len - offset; /* may want to use skb_headlen instead */
	
	kfree(wk); /* free the wk struct early */
	
#ifdef IP_FORWARD_ONLY_THROUGH_WQUEUES
	goto do_tx;
#endif /* IP_FORWARD_ONLY_THROUGH_WQUEUES */

	if (at_decoder)
	{
		if ((iph->protocol == IPPROTO_IPDEDUP_UDP) 
				|| (iph->protocol == IPPROTO_IPDEDUP_TCP))
		{
			rfs->total_diffs++;
			if (!pkt_patch(rfs, &skb, pkt_payload, offset, len))
			{
				/* TODO: initiate reset frame */
				goto free_skb;
			}
			rfs->recovered_diffs++;
			/* new skb means the offset, len, pkt_payload are all obsolete */
			iph = skb_ip_hdr(skb);
			pkt_payload = tcp_udp_payload(iph);
			offset = pkt_payload - skb->data;
			len = skb->len - offset; /* may want to use skb_headlen instead */
		}/* else fall through, regular packet */
	}
	/* acount for the total bytes after recovery */
	rfs->total_bytes += ntohs(iph->tot_len);
	
	add_skb_to_store(rfs, skb, offset, len);
	fingerprint_packet(rfs, pkt_payload, len);
	if (!at_decoder)
	{
#ifndef FAKE_DIFF /* keep the braces, otherwise if falls through */
		pkt_diffs(rfs, &skb, pkt_payload, offset, len);
		iph = skb_ip_hdr(skb); /* new skb, re-direct iph pointer */
#endif /* FAKE_DIFF */
	}
	add_fp_to_index(rfs);
	
#ifdef RESEQUENCE_IP_HDRS
	if (!at_decoder)
	{
		iph->check = 0;
		iph->check = ip_fast_csum(skb_network_header(skb), iph->ihl);
		skb->ip_summed = CHECKSUM_NONE; /* computed the checksum ourselves */
	}
#endif /* RESEQUENCE_IP_HDRS */
	
	/* if recovery was performed, skb now points to a forward-able packet */
	if (unlikely(do_route_me_harder(&skb)))
	{
		printk(DDUP_LOG": 'Hard' routing with no effect.\n");
		goto free_skb;
	}

#ifdef IP_FORWARD_ONLY_THROUGH_WQUEUES
	do_tx:
#endif /* IP_FORWARD_ONLY_THROUGH_WQUEUES */

	transmit_skb(skb);
	return;
	
	free_skb://
	kfree_skb(skb);
}

static inline int packet_match(struct iphdr *iph)
{
	int dest_port = 0;
	if ((iph->protocol == IPPROTO_UDP)
			|| (iph->protocol == IPPROTO_IPDEDUP_UDP))
	{
		struct udphdr *udph = ip_udp_hdr(iph);
		dest_port = ntohs(udph->dest);
	}
	else if ((iph->protocol == IPPROTO_TCP) || (iph->protocol
			== IPPROTO_IPDEDUP_TCP))
	{
		struct tcphdr *tcph = ip_tcp_hdr(iph);
		dest_port = ntohs(tcph->dest);
	}
	else
		return 0;
	
	return ((filter_port == 0) || (filter_port == dest_port));
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
static unsigned int dedup_nf_hook(unsigned int hook, struct sk_buff** pskb,
		const struct net_device *indev, const struct net_device *outdev,
		int(*okfn)(struct sk_buff*))
{
	struct sk_buff *skb = *pskb; /* peel off one layer of indirection */
#else
static unsigned int dedup_nf_hook(unsigned int hook, struct sk_buff* skb,
		const struct net_device *indev, const struct net_device *outdev,
		int (*okfn) (struct sk_buff*))
{
#endif

	/* netfilter hook on PF_INET, hence all traffic is IP */
	struct iphdr *iph = skb_ip_hdr(skb);
	int at_decoder = 0, pipeline_id;
	struct dedup_pipe *res = &pipe_state;
	
	BUG_ON(indev == NULL); /* on the forwarding path, hence indev != NULL */	
	/* be polite and don't interfere on the no_iface interface */
	if ((res->no_iface) && (
			(res->no_iface_ifindex == indev->ifindex) || 
			(outdev != NULL && res->no_iface_ifindex == outdev->ifindex) ))
	{
		ddup_rep.noiface++;
		goto out_fail;
	}
	
	/* it doesn't work on fragments */
	if (unlikely(iph->frag_off & ntohs(IP_OFFSET)))
	{
		ddup_rep.ipfrags++;
		goto out_fail;
	}
	
	/* logic only works on linear skbs (skb->data_len == 0) */
	if (unlikely(skb_linearize(skb)))
	{
		ddup_rep.skb_nonlinear++;
		printk(DDUP_LOG": Unable to linearize skb\n");
		goto out_fail;
	}
	
	if (likely(res->pipe_iface))
		at_decoder = res->pipe_iface_ifindex == indev->ifindex;
	else
		at_decoder = ((indev->name != NULL) && 
			(strcmp(indev->name, pipe_iface) == 0));

	if (likely(packet_match(iph)))
	{
		struct dedup_skb_work *skb_work =
#ifdef USE_KMEM_CACHE_WORK
				kmem_cache_alloc(dedup_skb_work_cache, GFP_ATOMIC);
#else
				kmalloc(sizeof(*skb_work), GFP_ATOMIC);		
#endif /* USE_KMEM_CACHE_WORK */

		if (unlikely(!skb_work))
		{
			printk(DDUP_LOG": Unable to allocate struct dedup_skb_work\n");
			goto out_fail;
		}
		
		INIT_LITE_WORK(&skb_work->work, dedup_skb_work_fn);
		skb_work->skb = skb;
		skb_work->at_decoder = at_decoder;
		
		/* this hash function has to be the same on the decoder
		 * use ip header identification field */
		
		if (!at_decoder)
		{
#ifdef RESEQUENCE_IP_HDRS
			iph->id = seq_inc_return();
#endif /* RESEQUENCE_IP_HDRS */
		}
		
		pipeline_id = iph->id % cores;
		skb_work->rfs = &rf_st[pipeline_id];
		fifo_queue_work_in_bh(wqueue[pipeline_id], &skb_work->work);

		if (at_decoder)
			ddup_rep.at_decoder++;
		else
			ddup_rep.at_encoder++;
		return NF_STOLEN;
	}
	out_fail: //
	ddup_rep.nomatch++;
	return NF_ACCEPT;
}

/* procfile function called to output data to userspace through /proc */
int proc_put_userland(char *page, char **start, off_t off, int count, 
		int *eof, void *data)
{
	int len = 0, i;
	char *dest = (char *) page;

	len += sprintf(dest, "Packets received on forbidden interface: %u\n",
			ddup_rep.noiface);
	dest = &(((char *) page)[len]);

	len += sprintf(dest, "Packets received at encoder: %u, at decoder: %u\n",
			ddup_rep.at_encoder, ddup_rep.at_decoder);
	dest = &(((char *) page)[len]);

	len += sprintf(dest, "Packets not processed: IP_FRAGMENT %u, no match %u, "
			"non linear skb %u\n", 
			ddup_rep.ipfrags, ddup_rep.nomatch, ddup_rep.skb_nonlinear);
		dest = &(((char *) page)[len]);

	for (i = 0; i < cores; i++)
	{
		len += sprintf(dest, "Pipeline %d:\n", i);
		dest = &(((char *) page)[len]);
		
		len += sprintf(dest, "\tDiff packets: %u patches, %u encoded, "
				"%u decoded, unrecoverable (%u collision, %u corrupted), "
				"%u total, [bytes saved: %lld / %llu.0]\n", rf_st[i].patches, 
				rf_st[i].encoded_diffs, rf_st[i].recovered_diffs, 
				rf_st[i].unrecoverable_diffs, rf_st[i].corrupted_diffs, 
				rf_st[i].total_diffs, rf_st[i].bytes_saved, rf_st[i].total_bytes);
		dest = &(((char *) page)[len]);
		
		len += sprintf(dest, "\tPacket store next_pkt_id: %u, capacity %u\n", 
				rf_st[i].next_pkt_id, rf_st[i].pkts.cnt);
		dest = &(((char *) page)[len]);

		len += sprintf(dest, "\tRepresentative fingerprints: %u / %u\n",
				rf_st[i].fp.repr_fp, rf_st[i].fp.fp);
		dest = &(((char *) page)[len]);

		len += sprintf(dest, "\tFingerprint index: capacity %u, "
				"updated %u, collisions %u, overwrite-invalid %u, "
				"same-fingerprint %u\n", rf_st[i].rfidx.cnt, 
				rf_st[i].rfidx.updates, rf_st[i].rfidx.collision,
				rf_st[i].rfidx.overwrite_invalid, 
				rf_st[i].rfidx.same_fingerprint);
		dest = &(((char *) page)[len]);
		
		len += sprintf(dest, "\tWork pending on queue: %lu\n", 
				wqueue[i]->items);
		dest = &(((char *) page)[len]);

	}
	
	PROC_READ_RETURN(page, start, off, count, eof, len);
}

/* procfile read data from userspace */
int proc_get_userland(struct file *filp, const char __user *buff,
		unsigned long count, void *data)
{
	if (!procfile_page)
		return count;

	if (count> PAGE_SIZE)
		return -EINVAL;

	if (copy_from_user(procfile_page, buff, count))
		return -EFAULT;

	if (strncmp(procfile_page, COUNTERS_RESET_STRING,
					strlen(COUNTERS_RESET_STRING)) == 0)
	memset(&ddup_rep, 0, sizeof(ddup_rep));

	return count;
}

static void do_exit_ipdedup(void)
{
	int i;
	if (dedup_nf_hook_ops.hook) /* hook function should have been set */
	{
		nf_unregister_hook(&dedup_nf_hook_ops);
		memset(&dedup_nf_hook_ops, 0, sizeof(dedup_nf_hook_ops));
	}

	if (wqueue)
	{
		for (i = 0; i < cores; i++)
			if (wqueue[i])
				destroy_fifo_wqueue(wqueue[i]);
		kfree(wqueue);
		wqueue = NULL;
	}

#ifdef USE_KMEM_CACHE_WORK
	if (dedup_skb_work_cache)
	{
		kmem_cache_destroy(dedup_skb_work_cache);
		dedup_skb_work_cache = NULL;
	}
#endif /* USE_KMEM_CACHE_WORK */
	
	if (proc_entry)
	{
		remove_proc_entry(procfile, NULL);
		proc_entry = NULL;
	}
	if (procfile_page)
	{
		free_page((unsigned long) procfile_page);
		procfile_page = NULL;
	}
	
	if (rf_st)
	{
		for (i = 0; i < cores; i++)
			destroy_rf_state(&(rf_st[i]));
		kfree(rf_st);
		rf_st = NULL;
	}
}

static int __init init_ipdedup(void)
{
	int err = 0, i;
	struct net_device *iface_dev;

	/* validate the module parameters */
	if ((cores < 1) || (cores > num_present_cpus()))
	{
		printk(DDUP_LOG": Cannot satisfy request for %d cores, "
				"max available: %d\n", cores, num_present_cpus());
		cores = SUPERSCALAR_FACTOR;
	}
	if ((mtu < 1) || (mtu > ETH_JUMBO_MTU))
		mtu = ETH_MTU; /* fallback value */
	if (rabin_window < 2 || rabin_window > mtu)
		rabin_window = DEFAULT_RABIN_WINDOW;

	/* the entire dedup pipe state */
	memset(&pipe_state, 0, sizeof(pipe_state));
	if (!mirror_ip)
	{
		printk(DDUP_LOG": Mirror IP address not given\n");
		err = -EINVAL;
		goto out;
	}
	pipe_state.remote.sin_family = AF_INET;
	pipe_state.remote.sin_addr.s_addr = in_aton(mirror_ip);
	pipe_state.remote.sin_port = htons(shard_udp_port);

	pipe_state.addr.sin_family = AF_INET;
	pipe_state.addr.sin_addr.s_addr = local_ip ? 
		in_aton(local_ip) : htonl(INADDR_ANY);
	pipe_state.addr.sin_port = htons(shard_udp_port);

	/* get interface indices for the pipe_iface and no_iface */
	if ((strcmp(no_iface, "") != 0)
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
		&& ((iface_dev = dev_get_by_name(no_iface)) != NULL))
#else
		&& ((iface_dev = dev_get_by_name(&init_net, no_iface)) != NULL))
#endif		
	{
		pipe_state.no_iface = 1;
		/* Are interface indices allowed to be negative? */
		pipe_state.no_iface_ifindex = iface_dev->ifindex;
		dev_put(iface_dev);
	}

	if ((strcmp(pipe_iface, "") != 0)
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
		&& ((iface_dev = dev_get_by_name(pipe_iface)) != NULL))
#else
		&& ((iface_dev = dev_get_by_name(&init_net, pipe_iface)) != NULL))
#endif		
	{
		pipe_state.pipe_iface = 1;
		/* Are interface indices allowed to be negative? */
		pipe_state.pipe_iface_ifindex = iface_dev->ifindex;
		dev_put(iface_dev);
	}

	if (!(rf_st = kmalloc(cores * sizeof(struct rf_state), GFP_KERNEL)))
	{
		printk(DDUP_LOG": Unable to allocate rf_state\n");
		err = -ENOMEM;
		goto out;
	}
	for (i = 0; i < cores; i++)
	{
		if ((err = init_rf_state(&(rf_st[i]), packet_store_count, mtu, 
				rf_index_count, rf_index_buckets, rabin_window, 
				rabin_representative_bits)) != 0)
		{
			printk(DDUP_LOG": Unable to allocate rf_state[%d]\n", i);
			goto out;
		}
	}
	
	if (!(procfile_page = (void *) get_zeroed_page(GFP_KERNEL)))
	{
		printk(DDUP_LOG": Unable to allocate procfile page\n");
		err = -ENOMEM;
		goto out;
	}
	memset(&ddup_rep, 0, sizeof(ddup_rep));
	/* create the proc file */
	if (!(proc_entry = create_proc_entry(procfile, 0666, NULL)))
	{
		printk(DDUP_LOG": Unable to create proc file %s\n", procfile);
		err = -EINVAL;
		goto out;
	}
	proc_entry->write_proc = proc_get_userland;
	proc_entry->read_proc = proc_put_userland;
	proc_entry->owner = THIS_MODULE;

#ifdef USE_KMEM_CACHE_WORK
	if (!(dedup_skb_work_cache = KMEM_CACHE(dedup_skb_work, 
			SLAB_HWCACHE_ALIGN | SLAB_PANIC)))
	{
		printk(DDUP_LOG": Unable to allocate kmem-cache for work_struct_lite\n");
		err = -ENOMEM;
		goto out;
	}
#endif /* USE_KMEM_CACHE_WORK */
	
	/* prep the workqueues, in fact use kernel threads bound to cores */
	if (!(wqueue = kmalloc(cores * sizeof(struct fifo_wqueue *), GFP_KERNEL)))
	{
		printk(DDUP_LOG": Unable to allocate wqueue\n");
		err = -ENOMEM;
		goto out;
	}
	for (i = 0; i < cores; i++)
	{
		char wqname[64];
		cpumask_t wqueue_cpu_mask = CPU_MASK_ALL; /* cpumask_of_cpu(0); */

#ifdef BIND_WORKQUEUES_TO_CPUS
		int cpus = num_present_cpus();
#ifdef BIND_WORKQUEUES_TO_CORES
		wqueue_cpu_mask = cpumask_of_cpu(logical_cpu_map[i % cpus]);
#else
		/* turns out CPU ids are already layed out for this case */
		wqueue_cpu_mask = cpumask_of_cpu(i % cpus);
#endif /* BIND_WORKQUEUES_TO_CORES */
#endif /* BIND_WORKQUEUES_TO_CPUS */

		sprintf(wqname, "pipeline-%d", i);
		if (!(wqueue[i] = create_fifo_wqueue(wqname, wqueue_cpu_mask, -5)))
		{
			err = -ENOMEM;
			goto out;
		}
	}
	
	/* the netfilter hook */
	memset(&dedup_nf_hook_ops, 0, sizeof(dedup_nf_hook_ops));
	dedup_nf_hook_ops.hook = dedup_nf_hook;
	dedup_nf_hook_ops.hooknum = hook_chain;
	dedup_nf_hook_ops.pf = PF_INET;
	dedup_nf_hook_ops.priority = hook_prio;
	err = nf_register_hook(&dedup_nf_hook_ops);

	out:
	if (err != 0)
	{
		do_exit_ipdedup();
		printk(DDUP_LOG": Unable to load IP packet dedup module\n");
	}
	else
	{
		printk(DDUP_LOG":\n");
		printk(DDUP_LOG": Loaded IP packet dedup module\n");
	}
	return err;
}

static void __exit exit_ipdedup(void)
{
	do_exit_ipdedup();
	printk(DDUP_LOG": Unloaded IP packet dedup module\n");
}

module_init( init_ipdedup);
module_exit( exit_ipdedup);

MODULE_DESCRIPTION("IP packet level redundancy elimination kernel module");
MODULE_AUTHOR("Tudor Marian");
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(a)
#endif
MODULE_LICENSE("Dual BSD/GPL");
