
/*
 * Maelstrom (NSDI'08) version 2 implementation.
 *
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 */

#include <linux/sched.h>
#include "util.h"
#include "wqueue.h"
#include "hashmap.h"
#include "mael.h"

#define BIND_WORKQUEUES_TO_CPUS
//#define BIND_WORKQUEUES_TO_CORES
#define USE_KMEM_CACHE_WORK
//#define RESEQUENCE_IP_HDRS
//#define NF_HOOK_OUT_SKB
//#define FORWARD_ONLY

#define SUPERSCALAR_FACTOR				4
#define NC								(-5) /* workqueue and task nice val */
#define IP_PACKET_CSUM					1
#define LOG								"MAELSTROM"
#define DEFAULT_KMALLOC_FLAGS			GFP_ATOMIC
#define COUNTERS_RESET_STRING			"reset-counters"
#define FEC_ENCODER_SND					"fec_snd_wq"
#define FEC_DECODER_RCV					"fec_rcv"
#define ONE_K							1024
#define ETH_MTU							1500
//#define UDP_RCV_CPU_ID					0

static int rcv_fec_udp_port = 6669;
module_param(rcv_fec_udp_port, int, 0000);
MODULE_PARM_DESC(rcv_fec_udp_port, "The port on which UDP FEC traffic travels");
static int xor_r = 8;
module_param(xor_r, int, 0000);
MODULE_PARM_DESC(xor_r, "The *R* parameter in the layered interleaving");
static int xor_c = 3;
module_param(xor_c, int, 0000);
MODULE_PARM_DESC(xor_c, "The *C* parameter in the layered interleaving");
static int cores = SUPERSCALAR_FACTOR;
module_param(cores, int, 0000);
MODULE_PARM_DESC(cores, "The superscalar factor (number of CPUs/cores)");
static int recovery_table_len = 1024;
module_param(recovery_table_len, int, 0000);
MODULE_PARM_DESC(recovery_table_len, 
		"The size of the recovery data hashtable");
static int recovery_table_bin_capacity = 4;
module_param(recovery_table_bin_capacity, int, 0000);
MODULE_PARM_DESC(recovery_table_bin_capacity, 
		"The capacity of the recovery data hashtable");
static int fec_recovery_table_len = 1024;
module_param(fec_recovery_table_len, int, 0000);
MODULE_PARM_DESC(fec_recovery_table_len, 
		"The size of the fec recovery data hashtable");
static int fec_recovery_table_bin_capacity = 4;
module_param(fec_recovery_table_bin_capacity, int, 0000);
MODULE_PARM_DESC(fec_recovery_table_bin_capacity, 
		"The capacity of the fec recovery data hashtable");
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 24)
static int hook_chain = NF_IP_FORWARD;
#else
static int hook_chain = NF_INET_FORWARD;
#endif
module_param(hook_chain, int, 0000);
MODULE_PARM_DESC(hook_chain, "Netfilter hook chain");
static int hook_prio = NF_IP_PRI_FIRST;
module_param(hook_prio, int, 0000);
MODULE_PARM_DESC(hook_prio, "Netfilter hook priority");
static int filter_port = 0; /* 0 means ALL traffic is FECed */
module_param(filter_port, int, 0000);
MODULE_PARM_DESC(filter_port, "The port on which traffic is intercepted on, "
		"0 means all but UDP FEC port");
static char* no_iface = "";
module_param(no_iface, charp, 0000);
MODULE_PARM_DESC(no_iface, "Interface that should not be interfered with");
static char* pipe_iface = "";
module_param(pipe_iface, charp, 0000);
MODULE_PARM_DESC(pipe_iface, "Interface on which incoming packets on the "
		"filter port should be buffered for FEC");
static char *procfile = "maelstrom";
module_param(procfile, charp, 0000);
MODULE_PARM_DESC(procfile, "The name of the /proc/<file>");
static char *mirror_ip = NULL;
module_param(mirror_ip, charp, 0000);
MODULE_PARM_DESC(mirror_ip, "The IP address of the mirroring host");
static char *local_ip = NULL;
module_param(local_ip, charp, 0000);
MODULE_PARM_DESC(local_ip, "The IP address of the mirroring host");

int logical_cpu_map[] = {0, 8, 4, 12, 1, 9, 5, 13, 2, 10, 6, 14, 3, 11, 7, 15};
void *procfile_page = NULL;
static inline int rcv_fec_cmd_port(void)	{ return (rcv_fec_udp_port+1); }
int slot_len = (2*ONE_K); /* sufficient for regular MTU packets & FEC headers */

static int interleaves_cnt = 9;
/* interleaves[0] must be == 1 */
static int interleaves[] = {1, 19, 41, 83, 167, 337, 673, 1361, 2729};
static inline int get_interleave(int i){ return interleaves[i]; }

#ifdef USE_KMEM_CACHE_WORK
struct kmem_cache *mael_skb_work_cache = NULL;
#ifndef KMEM_CACHE
#define KMEM_CACHE(__struct, __flags) kmem_cache_create(#__struct,\
				sizeof(struct __struct), __alignof__(struct __struct),\
				(__flags), NULL)
#endif /* KMEM_CACHE */
#endif

static inline void *default_alloc(unsigned long size)
{ 
	return kmalloc(size, DEFAULT_KMALLOC_FLAGS);
}

struct mael_reports
{
	unsigned int ipfrags;
	unsigned int noiface;
	unsigned int at_encoder;
	unsigned int at_decoder;
	unsigned int nomatch;
	unsigned int skb_nonlinear;
};
struct mael_reports mael_rep;

struct fec_bin
{
	void *start_buf;
	void *buf;
	int r;
	size_t len;
	size_t size;
};

/* ENCODER FEC BINS */
struct fec_bins
{
	struct fec_bin ***bin;
	unsigned long packets; /* the number of packets passing through */
	int xor_c, xor_r;
	struct fec_bin **ret; /* used to return full layers of FEC */
};

/* DECODER  */
struct fec_recovery
{
	struct hashmap packets;
	struct hashmap fec_packets;
};

struct mael_stage_resources
{
	struct fec_bins bins;
	struct fec_recovery recovery_tables;
	struct fifo_wqueue *wq;
	struct socket *fec_snd_sock; /* UDP FEC send socket */
	struct socket *raw_sock; /* raw socket used to send recovered packets */
	/* stat counters */
	uint32_t ip_tx, fec_tx, fec_tx_fail, ip_rx, fec_rx, fec_tx_recovered, 
		fec_rx_failed_recovery, fec_rx_not_needed, collision_cnt;
};

struct mael_resources
{
	int pipe_iface, pipe_iface_ifindex, no_iface, no_iface_ifindex, 
		rcv_bufsize, stages;
	struct mael_stage_resources *mael_stages;
	struct sockaddr_in addr, remote;
	struct task_struct *fec_decoder_rcv;
	struct completion fec_decoder_rcv_done, fec_decoder_rcv_started;
	struct nf_hook_ops nf_hook;
	struct proc_dir_entry *proc_entry;	
};
struct mael_resources mael_res;

#define fec_hdr(r)	(sizeof(struct xorhdr) + r * sizeof(struct maelhdr))
#define bin_total_len(bin, r)	(fec_hdr(r) + bin->len)

struct xorhdr
{
	short id;
	int cnt;
	size_t len;
};

struct maelhdr
{
	__u32 src, dst;
	__u16 id, offset;
	__u32 csum;
	size_t datasize;
};

unsigned long maelhdr_hash(void *key)
{
	struct maelhdr *mh = key;
	return (mh->src ^ mh->dst ^ mh->id ^ 
			((long) (mh->offset << 16)) ^ mh->csum ^ mh->datasize);	
}

int maelhdr_cmp(void *key1, void *key2)
{
	struct maelhdr *mh1 = key1;
	struct maelhdr *mh2 = key2;
	int eq = 
		(	mh1->src		== 	mh2->src && 
			mh1->dst 		== 	mh2->dst && 
			mh1->id 		==	mh2->id &&
			mh1->offset 	== 	mh2->offset &&
			mh1->csum		==	mh2->csum &&
			mh1->datasize	==	mh2->datasize);
	return !(eq);
}

struct recovery_packet
{
	struct maelhdr h;
	struct hash_head hhead;
	struct iphdr *iph; /* IP packet starting w/ the header */
};

void recovery_table_overflow(struct hash_head *entry)
{
	struct recovery_packet *rp = container_of(entry, struct recovery_packet, 
		hhead);
	kfree((void *) rp->iph);
	kfree(rp);
}

void fec_recovery_table_overflow(struct hash_head *entry) {}

static void free_recovery_tables(struct fec_recovery *tables)
{
	hashmap_release(&tables->packets, recovery_table_overflow, kfree);
	hashmap_release(&tables->fec_packets, fec_recovery_table_overflow, kfree);
	memset(tables, 0, sizeof(struct fec_recovery));
}

static int init_recovery_tables(struct fec_recovery *tables)
{
	int err = 0;
	memset(tables, 0, sizeof(struct fec_recovery));
	
	if (recovery_table_len < 1 || fec_recovery_table_len < 1)		
	{
		printk(LOG": Invalid recovery hashtable capacity\n");
		err = -EINVAL;
		goto out_err;
	}
	
	if ((err = init_hashmap(&tables->packets, recovery_table_len, 
			recovery_table_bin_capacity, default_alloc,
			maelhdr_hash, maelhdr_cmp, recovery_table_overflow)) < 0)
	{
		printk(LOG": Unable to initialize recovery tables, error %d\n", 
				err);
		goto out_err;
	}
	
	if ((err = init_hashmap(&tables->fec_packets, fec_recovery_table_len, 
			fec_recovery_table_bin_capacity, default_alloc,
			maelhdr_hash, maelhdr_cmp, fec_recovery_table_overflow)) < 0)
	{
		printk(LOG": Unable to initialize fec recovery tables, error "
				"%d\n", err);
		goto out_err;
	}
	
	return err;
	out_err:
	free_recovery_tables(tables);
	return err;
}

int init_fec_bin(struct fec_bins *bins, struct fec_bin *bin, 
		size_t bin_size, int kmalloc_flags)
{
	/* MUST be zeroed out */
	if (!(bin->start_buf = kzalloc(bin_size, kmalloc_flags)))
	{
		printk(LOG": Unable to init fec bin\n");
		return -ENOMEM;
	}
	bin->buf = bin->start_buf + fec_hdr(bins->xor_r); /* header room */
	bin->size = bin_size;
	bin->r = 0;
	bin->len = 0;
	return 0;
}

size_t ip_packet_to_mhdr(struct maelhdr *mhdr, struct iphdr* iph)
{
	
	mhdr->src = iph->saddr;
	mhdr->dst = iph->daddr;
	mhdr->id = ntohs(iph->id);
	mhdr->offset = ntohs(iph->frag_off) & IP_OFFSET;
	mhdr->datasize = ntohs(iph->tot_len);
#ifdef IP_PACKET_CSUM
	/* the checksum is only of data payload, TTL in the IP header interferes */
	mhdr->csum = csum_partial((char *) (ip_payload(iph)), 
		mhdr->datasize - (iph->ihl * 4), ~0);
#else
	mhdr->csum = 0; /* otherwise the TTL field interferes */
#endif
	return mhdr->datasize;
}

static int init_bins(struct fec_bins *bins, int xor_r, int xor_c)
{
	int i, j, flags = GFP_KERNEL;

	memset(bins, 0, sizeof(*bins));

	bins->xor_r = xor_r;
	bins->xor_c = xor_c;
	
	if (!(bins->bin = kzalloc(bins->xor_c * sizeof(struct fec_bin **), flags)))
	{
		printk(LOG": Unable to allocate bin\n");
		return -ENOMEM;
	}
	
	for (i = 0; i < bins->xor_c; i++)
	{
		if (!(bins->bin[i] = kzalloc(get_interleave(i) * 
				sizeof(struct fec_bin *), flags)))
		{
			printk(LOG": Unable to allocate bins[]\n");
			return -ENOMEM;
		}
		for (j = 0; j < get_interleave(i); j++)
		{
			if (!(bins->bin[i][j] = kmalloc(sizeof(struct fec_bin), flags)))
			{
				printk(LOG": Unable to allocate bins[][]\n");
				return -ENOMEM;
			}
			if (init_fec_bin(bins, bins->bin[i][j], slot_len, flags) != 0)
			{
				printk(LOG": Unable to allocate fec bin\n");
				return -ENOMEM;
			}
		}
	}
	
	/* don't forget about the ret array of pointers */
	if (!(bins->ret = kzalloc(bins->xor_c * sizeof(struct fec_bin *), flags)))
	{
		printk(LOG": Unable to allocate bins->ret[]");
		return -ENOMEM;
	}
	
	return 0;
}

void free_bins(struct fec_bins *bins)
{
	int i, j;
	if (bins->bin)
	{
		for (i = 0; i < bins->xor_c; i++)
		{
			if (bins->bin[i])
			{
				for (j = 0; j < get_interleave(i); j++)
				{
					if (bins->bin[i][j])
					{
						if (bins->bin[i][j]->start_buf)
						{
							kfree(bins->bin[i][j]->start_buf);
							memset(bins->bin[i][j], 0, sizeof(struct fec_bin));
						}
						kfree(bins->bin[i][j]);
						bins->bin[i][j] = NULL;
					}
				}
				kfree(bins->bin[i]);
				bins->bin[i] = NULL;
			}
		}
		kfree(bins->bin);
		bins->bin = NULL;
	}
	
	if (bins->ret)
	{
		kfree(bins->ret);
		bins->ret = NULL;
	}
	memset(bins, 0, sizeof(*bins));
}

/* 
 * Returns the number of slots ready for sending, and in the bins->ret the 
 * pointers to the ready bins; bins->ret is an array of maximum size (xor_c)
 *
 * Should probably use @skb_xor_bits
 */
int fec_packet(struct fec_bins *bins, struct iphdr *packet, size_t len)
{
	int i = 0, ready_bins = 0;
	
	for (i = 0; i < bins->xor_c; i++)
	{
		int binidx = bins->packets % get_interleave(i);
		struct fec_bin *bin = bins->bin[i][binidx];
		struct maelhdr *mhdr = bin->start_buf + sizeof(struct xorhdr);
		
		ip_packet_to_mhdr(&mhdr[bin->r], packet); /* bin->r is the next slot */
		bin->r++;
		
		if (bin->r == 1) /* first packet */
		{
			memcpy(bin->buf, (void *) packet, len);
			 /* if the old slot was larger than new slot, zero out tail */
			if (bin->len > len) /* at init time (bin->len == 0) */
				memset(bin->buf+len, 0, (bin->len - len));
			bin->len = len;
		}
		else
		{
			xor_bytes(bin->buf, (void *) packet, len);
			if (len > bin->len)
				bin->len = len;
			if (bin->r == bins->xor_r)
			{
				struct xorhdr *xhdr = (struct xorhdr *) bin->start_buf;
				xhdr->cnt = bins->xor_r;
				xhdr->len = bin_total_len(bin, bins->xor_r);
				
				bins->ret[ready_bins] = bin;
				ready_bins ++;
				bin->r = 0;
			}
		}
	}
	bins->packets++;
	return ready_bins;
}

struct mael_skb_work
{
	struct work_struct_lite work;
	union
	{
		struct sk_buff *skb;
		struct kvec buf;
	} packet;
	struct mael_stage_resources *msr;
	short stage_id, is_fec;
};

/* 
 * the invariant is that there is a single task (context of execution)
 * handling a FEC encode / decode stage. The same task will handle both
 * stages --- there may be another oportunity to parallelize that further.
 */
static void ip_packet_at_encoder(struct work_struct_lite *work)
{
	struct mael_skb_work *wk = container_of(work, struct mael_skb_work, work);
	struct sk_buff *skb = wk->packet.skb;
	struct iphdr *iph = skb_ip_hdr(skb);
	struct mael_stage_resources *res = wk->msr;
	short stage_id = wk->stage_id;
	int fecs_to_send = 0, i, len = skb->len;
	
	kfree(wk); /* release early */
	
#ifdef FORWARD_ONLY
	transmit_skb(skb);
	res->ip_tx++;
	return;
#endif //FORWARD_ONLY
	
	/* fec the packet */
	fecs_to_send = fec_packet(&res->bins, iph, len);
	
	/* forward original skb */
	transmit_skb(skb);
	res->ip_tx++;
	
	/* FEC packet and transmit any well formed redundant packet */
	for (i = 0; i < fecs_to_send; i++)
	{
		int err = 0;
		struct fec_bin *bin = res->bins.ret[i];
		struct xorhdr *xhdr = (struct xorhdr *) bin->start_buf;
		
		xhdr->id = stage_id; /* Warning, do not forget to set this field */
		err = ksock_snd(res->fec_snd_sock, &mael_res.remote, 
			(void *) xhdr, xhdr->len);
		if (unlikely(err < 0))
		{
			printk(LOG": Unable to send UDP packet to " NIPQUAD_FMT ":%d\n", 
				NIPQUAD(mael_res.remote.sin_addr.s_addr), 
				ntohs(mael_res.remote.sin_port));
			res->fec_tx_fail++;
		}
		else
			res->fec_tx++;
	}
}

static int add_packet_to_recovery_map(struct mael_stage_resources *res, 
	struct iphdr *iph)
{
	int err = 0;
	size_t len;
	struct recovery_packet *rp;
	struct hash_head *collision;
	
	if(unlikely(!(rp = default_alloc(sizeof(*rp)))))
	{
		printk(LOG": Unable to kmalloc struct recovery_packet\n");
		err = -ENOMEM;
		goto out;
	}
	
	len = ip_packet_to_mhdr(&rp->h, iph);
	
	if (unlikely(!(rp->iph = default_alloc(len))))
	{
		printk(LOG": Unable to kmalloc skb copy\n");
		if (rp)
			kfree(rp);
		err = -ENOMEM;
		goto out;
	}
	
	memcpy(rp->iph, iph, len);
	rp->hhead.key = &rp->h; /* set the key */
		
	collision = put_item(&res->recovery_tables.packets, &rp->hhead);
	if (collision != NULL)
	{
		recovery_table_overflow(collision);
		res->collision_cnt++;
	}
	
	out:
	return err;
}

static void ip_packet_at_decoder(struct work_struct_lite *work)
{
	struct mael_skb_work *wk = container_of(work, struct mael_skb_work, work);
	struct mael_stage_resources *res = wk->msr;
	int is_fec = (int) wk->is_fec;
	
#ifdef FORWARD_ONLY
	if (is_fec)
		kfree(wk->packet.buf.iov_base);
	else
	{
		res->ip_rx++;
		transmit_skb(wk->packet.skb);
	}
	kfree(wk);
	return;
#endif //FORWARD_ONLY
	
	/* implement simplified version first */
	if (is_fec)
	{
		int i, missing_packets = 0, raw_sent;
		//size_t len = wk->packet.buf.iov_len; /* don't actually need this */
		void *packet_payload, *buf = wk->packet.buf.iov_base;
		struct xorhdr *xh = (struct xorhdr *) buf;
		struct maelhdr *mh = buf + sizeof(struct xorhdr);
		
		kfree(wk); /* release early */
		
		res->fec_rx++;
		
		/* perform simple recovery */
		for (i = 0; i < xh->cnt; i++)
			if (!contains_key(&res->recovery_tables.packets, (void *) &mh[i]))
				missing_packets++;
		
		packet_payload = buf + fec_hdr(xh->cnt);
		switch (missing_packets)
		{
		case 0:
			res->fec_rx_not_needed++;
			break;
		case 1:
			for (i = 0; i < xh->cnt; i++)
			{
				struct hash_head *item = get_item(
					&res->recovery_tables.packets, &mh[i]);
				if (item)
				{
					struct recovery_packet *rp = container_of(item, 
						struct recovery_packet, hhead);
					xor_bytes(packet_payload, (void *)rp->iph, mh[i].datasize);
				}
			}
			/* forward recovered packet, if any */
			raw_sent = raw_send(res->raw_sock, (struct iphdr *) packet_payload);
			if (unlikely(raw_sent < 0))
			{
				printk(LOG": Unable to send raw bytes %d\n", raw_sent);
				res->fec_rx_failed_recovery++;
				break;
			}
			res->fec_tx_recovered++;
			
			/* store the recovered packet */
			add_packet_to_recovery_map(res, packet_payload);
			
			break;
		default:
			res->fec_rx_failed_recovery++;
			break;
		}
		
		kfree(buf);
	}
	else
	{
		struct sk_buff *skb = wk->packet.skb;
		struct iphdr *iph = skb_ip_hdr(skb);

		kfree(wk); /* release early */
		
		res->ip_rx++;
		/* add packet to recovery map */
		add_packet_to_recovery_map(res, iph);
		/* forward original skb */
		transmit_skb(skb);
	}
}

static inline int packet_match(struct iphdr *iph)
{
	int dest_port = 0;
	if (likely((iph->protocol == IPPROTO_TCP) || 
			(iph->protocol == IPPROTO_UDP)))
	{
		if (iph->protocol == IPPROTO_UDP)
		{
			struct udphdr *udph = ip_udp_hdr(iph);
			dest_port = ntohs(udph->dest);

			/* careful abou the FEC and the SMFS port */
			if (rcv_fec_udp_port == dest_port || 
				rcv_fec_cmd_port() == dest_port)
				return 0;
		}
		else if (iph->protocol == IPPROTO_TCP)
		{
			struct tcphdr *tcph = ip_tcp_hdr(iph);
			dest_port = ntohs(tcph->dest);
		}		
	}
	else
		return 0;
	return ((filter_port == 0) || (filter_port == dest_port));
}

#ifdef RESEQUENCE_IP_HDRS
/* If NAPI is used, then we are safe to increment this value, otherwise have
 * to provide an atomic_t like implementation for the short datatype. */
static unsigned short packet_id_cnt = 0;
/* wraps around when overflowing */
static inline unsigned short seq_inc_return(void) { return (packet_id_cnt++); }
#endif /* RESEQUENCE_IP_HDRS */

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
static unsigned int maelstrom_nf_hook(unsigned int hook,
		struct sk_buff** pskb,
		const struct net_device *indev,
		const struct net_device *outdev,
		int (*okfn) (struct sk_buff*))
{
	struct sk_buff *skb = *pskb; // removed one layer of indirection
#else
static unsigned int maelstrom_nf_hook(unsigned int hook,
		struct sk_buff* skb,
		const struct net_device *indev,
		const struct net_device *outdev,
		int (*okfn) (struct sk_buff*))
{
#endif
	
	/* netfilter hook on PF_INET, hence all traffic is IP */
	struct iphdr *iph = skb_ip_hdr(skb);
	struct mael_resources *res = &mael_res;
	int at_decoder = 0;
	
	BUG_ON(indev == NULL); /* on the forwarding path, hence indev != NULL */
	/* be polite and don't interfere on the no_iface interface */
	if ((res->no_iface) && ( (res->no_iface_ifindex == indev->ifindex) || 
			(outdev != NULL && res->no_iface_ifindex == outdev->ifindex) ))
	{
		mael_rep.noiface++;
		goto out_fail;
	}
	/* maelstrom doesn't work on fragments */
	if (unlikely(iph->frag_off & ntohs(IP_OFFSET)))
	{
		mael_rep.ipfrags++;
		goto out_fail;
	}
	
	/* I was lazy so XOR logic only works on linear skbs (skb->data_len == 0) */
	if (unlikely(skb_linearize(skb)))
	{
		mael_rep.skb_nonlinear++;
		printk(LOG": Unable to linearize skb\n");
		goto out_fail;
	}
	
	if (likely(res->pipe_iface))
		at_decoder = res->pipe_iface_ifindex == indev->ifindex;
	else
		at_decoder = ((indev->name != NULL) && 
			(strcmp(indev->name, pipe_iface) == 0));

	if (likely(packet_match(iph)))
	{
		struct mael_skb_work *skb_work =
#ifdef USE_KMEM_CACHE_WORK
			kmem_cache_alloc(mael_skb_work_cache, GFP_ATOMIC);
#else
			kmalloc(sizeof(*skb_work), GFP_ATOMIC);		
#endif /* USE_KMEM_CACHE_WORK */

		if (unlikely(!skb_work))
		{
			printk(LOG": Unable to allocate struct mael_skb_work\n");
			goto out_fail;
		}

		if (at_decoder)
		{
			INIT_LITE_WORK(&skb_work->work, ip_packet_at_decoder);
			mael_rep.at_decoder++;
			skb_work->is_fec = 0;
		}
		else
		{
			INIT_LITE_WORK(&skb_work->work, ip_packet_at_encoder);
			mael_rep.at_encoder++;
#ifdef RESEQUENCE_IP_HDRS
			iph->id = seq_inc_return();
#endif /* RESEQUENCE_IP_HDRS */
		}
		
		skb_work->packet.skb = skb;
		skb_work->stage_id = iph->id % res->stages;
		skb_work->msr = &res->mael_stages[skb_work->stage_id];
		fifo_queue_work_in_bh(skb_work->msr->wq, &skb_work->work);
		
		return NF_STOLEN;
	}
	
	out_fail:
	mael_rep.nomatch++;
	return NF_ACCEPT;
}

int fec_decoder_rcv_runner(void *data)
{
	struct mael_resources *res = (struct mael_resources *) data;
	int err = 0, size;
	struct socket *fec_sock = NULL;
	struct sockaddr_in sender_addr;
	
	memset(&sender_addr, 0, sizeof(sender_addr));
	
	current->flags |= PF_NOFREEZE;
	daemonize(FEC_DECODER_RCV);
	allow_signal(SIGKILL); /* ksocket_receive preempted by signals */
	
	/* both insmod and rmmod will have to wait for this */
	complete_all(&res->fec_decoder_rcv_started);
	
	/* prepare the UDP fec socket */
	if ((err = sock_create_kern(AF_INET, SOCK_DGRAM, IPPROTO_UDP, &fec_sock))<0)
	{
		printk(LOG": Unable to create UDP fec socket, errno %d\n", err);
		goto out;
	}
	if ((err = fec_sock->ops->bind(fec_sock, 
			(struct sockaddr *) &res->addr, sizeof(struct sockaddr))) < 0)
	{
		printk(LOG": Unable to bind UDP fec socket error %d\n", err);
		goto out;
	}

	while(!signal_pending(current))
	{
		void *recv_buf = NULL;
		if (unlikely(!(recv_buf = kmalloc(res->rcv_bufsize, GFP_ATOMIC))))
		{
			printk(LOG": Unable to allocate UDP fec receive buffer\n");
			err = -ENOMEM;
			goto out;
		}
		
		size = ksock_rcv(fec_sock, &sender_addr, recv_buf, res->rcv_bufsize);
		if (unlikely(size < 0))
		{
			if (size != -ERESTARTSYS) /* did not receive SIGKILL */
				printk(LOG": Error receiving UDP fec, sock_recvmsg "
					"error = %d\n", size);
		}
		else
		{
			struct xorhdr *xh = (struct xorhdr *) recv_buf;
			struct mael_skb_work *skb_work =
#ifdef USE_KMEM_CACHE_WORK
				kmem_cache_alloc(mael_skb_work_cache, GFP_ATOMIC);
#else
				kmalloc(sizeof(*skb_work), GFP_ATOMIC);		
#endif /* USE_KMEM_CACHE_WORK */

			if (unlikely(!skb_work))
			{
				printk(LOG": Unable to allocate struct mael_skb_work\n");
				goto cleanup;
			}
			
			INIT_LITE_WORK(&skb_work->work, ip_packet_at_decoder);
			skb_work->packet.buf.iov_base = recv_buf;
			skb_work->packet.buf.iov_len = size;
			skb_work->is_fec = 1;
			skb_work->stage_id = xh->id % res->stages;
			skb_work->msr = &res->mael_stages[skb_work->stage_id];
			fifo_queue_work(skb_work->msr->wq, &skb_work->work);
			
			recv_buf = NULL; /* mark buffer NULL to prevent reclaiming */
		}
		
		cleanup:
		if (recv_buf) /* cleanup */
		{
			kfree(recv_buf);
			recv_buf = NULL;
		}
	}
	
	out:
	if (fec_sock)
	{
		sock_release(fec_sock);
		fec_sock = NULL;
	}
	complete(&res->fec_decoder_rcv_done);
	return err;
}

int proc_put_userland(char *page, char **start, off_t off, int count, 
		int *eof, void *data)
{
	int len = 0, i;
	char *dest = (char *) page;

	len += sprintf(dest, "No iface: %u\n",mael_rep.noiface);
	dest = &(((char *) page)[len]);

	len += sprintf(dest, "Packets received at encoder: %u, at decoder: %u\n",
			mael_rep.at_encoder, mael_rep.at_decoder);
	dest = &(((char *) page)[len]);

	len += sprintf(dest, "Packets not processed: IP_FRAGMENT %u, no match %u, "
			"non linear skb %u\n", 
			mael_rep.ipfrags, mael_rep.nomatch, mael_rep.skb_nonlinear);
		dest = &(((char *) page)[len]);

	for (i = 0; i < mael_res.stages; i++)
	{
		struct mael_stage_resources *res = &mael_res.mael_stages[i];

		len += sprintf(dest, "Stage %d:\n", i);
		dest = &(((char *) page)[len]);
		
		len += sprintf(dest, "\tip_tx: %u, fec_tx: %u, fec_tx_fail: %u, "
			"ip_rx: %u, fec_rx: %u, fec_tx_recovered: %u, "
			"fec_rx_failed_recovery: %u, fec_rx_not_needed: %u, "
			"rx_collisions: %u\n", res->ip_tx, 
			res->fec_tx, res->fec_tx_fail, res->ip_rx, res->fec_rx, 
			res->fec_tx_recovered, res->fec_rx_failed_recovery, 
			res->fec_rx_not_needed, res->collision_cnt);
		dest = &(((char *) page)[len]);
	}
	
	PROC_READ_RETURN(page, start, off, count, eof, len);
}

/* procfile read data from userspace */
int proc_get_userland(struct file *filp, const char __user *buff,
		unsigned long count, void *data)
{
	if (!procfile_page)
		return count;

	if (count> PAGE_SIZE)
		return -EINVAL;

	if (copy_from_user(procfile_page, buff, count))
		return -EFAULT;

	if (strncmp(procfile_page, COUNTERS_RESET_STRING,
					strlen(COUNTERS_RESET_STRING)) == 0)
	memset(&mael_rep, 0, sizeof(mael_rep));

	return count;
}

/* tear down the maelstrom resources */
void free_mael_resources(struct mael_resources *res)
{
	int err = 0, i = 0;
	
	if (res->nf_hook.hook) /* hook function should have been set */
	{
		nf_unregister_hook(&res->nf_hook);
		memset(&res->nf_hook, 0, sizeof(struct nf_hook_ops));
	}
	
#ifdef USE_KMEM_CACHE_WORK
	if (mael_skb_work_cache)
	{
		kmem_cache_destroy(mael_skb_work_cache);
		mael_skb_work_cache = NULL;
	}
#endif /* USE_KMEM_CACHE_WORK */
	
	if (res->fec_decoder_rcv)
	{
		wait_for_completion(&res->fec_decoder_rcv_started);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 24)
		err = kill_proc(res->fec_decoder_rcv->pid, SIGKILL, 1);
#else
		err = kill_pid(find_pid_ns(res->fec_decoder_rcv->pid, &init_pid_ns), 
				SIGKILL, 1);
#endif
		if (err < 0)
			printk(LOG": Unable to kill fec receiver, error %d\n", err);
		else
			wait_for_completion(&res->fec_decoder_rcv_done);
		res->fec_decoder_rcv = NULL;
	}
	
	if (res->mael_stages)
	{
		for (i = 0; i < res->stages; i++)
		{
			struct mael_stage_resources *stage_res = &res->mael_stages[i];
			if (stage_res->wq)
			{
				destroy_fifo_wqueue(stage_res->wq);
				stage_res->wq = NULL;
			}
			if (stage_res->fec_snd_sock)
			{
				sock_release(stage_res->fec_snd_sock);
				stage_res->fec_snd_sock = NULL;
			}
			if (stage_res->raw_sock)
			{
				sock_release(stage_res->raw_sock);
				stage_res->raw_sock = NULL;
			}
			
			free_recovery_tables(&stage_res->recovery_tables);
			free_bins(&stage_res->bins);
		}
		kfree(res->mael_stages);
		res->mael_stages = NULL;
	}
	
	if (res->proc_entry)
	{
		remove_proc_entry(procfile, NULL);
		res->proc_entry = NULL;
	}
	if (procfile_page)
	{
		free_page((unsigned long) procfile_page);
		procfile_page = NULL;
	}
	
	memset(res, 0, sizeof(struct mael_resources));
}

struct init_mael_stage_struct
{
	struct work_struct_lite work;
	struct mael_stage_resources *res;
	struct sockaddr_in *local_addr;
	int ret; /* the return code */
};

/* raw sockets only like the IP_HDRINCL option set on a task != insmod */
void init_mael_stage_resources(struct work_struct_lite *work)
{
	int on = 1, err = 0;
	struct sockaddr_in local_addr;
	struct init_mael_stage_struct *stage_res = container_of(work, 
		struct init_mael_stage_struct, work);
	struct mael_stage_resources *res = stage_res->res;
	
	/* prepare the fec_snd_socket */
	if ((err = sock_create_kern(AF_INET, SOCK_DGRAM, IPPROTO_UDP, 
			&res->fec_snd_sock)) < 0)
	{
		printk(LOG": Unable to create UDP fec socket, errno %d\n", err);
		goto out;
	}
	
	memcpy(&local_addr, stage_res->local_addr, sizeof(local_addr));
	local_addr.sin_port = 0; /* bind to any port, but proper address */
	if ((err = res->fec_snd_sock->ops->bind(res->fec_snd_sock, 
			(struct sockaddr *) &local_addr, sizeof(struct sockaddr))) < 0)
	{
		printk(LOG": Unable to bind UDP fec_snd socket error %d\n", err);
		goto out;
	}
	
	/* prepare the raw socket */
	if ((err = sock_create_kern(PF_INET, SOCK_RAW, IPPROTO_RAW, 
			&res->raw_sock)) < 0)
	{
		printk(LOG": Unable to create RAW socket, error %d\n", err);
		goto out;
	}

	if ((err = res->raw_sock->ops->setsockopt(res->raw_sock, 
			IPPROTO_IP, IP_HDRINCL, (void *) &on, sizeof(on))) < 0)
	{
		printk(LOG": Unable to set the IP_HDRINCL for the raw socket, "
				"error %d\n", err);
		goto out;
	}
	
	out:
	stage_res->ret = err;
}

int init_mael_resources(struct mael_resources *res, int cores, int recvsz, 
	int r, int c)
{
	int err = 0, i;
	struct net_device *iface_dev;
	struct fifo_wqueue *bootstrap_wq = NULL;
	cpumask_t cpu_mask;
	
	memset(res, 0, sizeof(struct mael_resources));
	res->stages = cores;
	res->rcv_bufsize = recvsz;
	
	if (c >= interleaves_cnt)
	{
		c = interleaves_cnt-1;
		printk(LOG": WARNING, c parameter too large, max allowed is %d\n", c);
	}
	
	/* prepare the remote address */
	if (!mirror_ip)
	{
		printk(LOG": Mirror IP address not given\n");
		err = -EINVAL;
		goto out;
	}
	res->remote.sin_family = AF_INET;
	res->remote.sin_addr.s_addr = in_aton(mirror_ip);
	res->remote.sin_port = htons(rcv_fec_udp_port);

	res->addr.sin_family = AF_INET;
	if (local_ip)
		res->addr.sin_addr.s_addr = in_aton(local_ip);
	else
		res->addr.sin_addr.s_addr = htonl(INADDR_ANY);
	res->addr.sin_port = htons(rcv_fec_udp_port);
	
	/* get interface indices for the pipe_iface and no_iface */
	if ((strcmp(no_iface, "") != 0) 
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
			&& ((iface_dev = dev_get_by_name(no_iface)) != NULL))
#else
			&& ((iface_dev = dev_get_by_name(&init_net, no_iface)) != NULL))
#endif		
	{
		res->no_iface = 1;
		/* Are interface indices allowed to be negative? */
		res->no_iface_ifindex = iface_dev->ifindex;
		dev_put(iface_dev);			
	}

	if ((strcmp(pipe_iface, "") != 0) 
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
			&& ((iface_dev = dev_get_by_name(pipe_iface)) != NULL))
#else
			&& ((iface_dev = dev_get_by_name(&init_net, pipe_iface)) != NULL))
#endif		
	{
		res->pipe_iface = 1;
		/* Are interface indices allowed to be negative? */
		res->pipe_iface_ifindex = iface_dev->ifindex;
		dev_put(iface_dev);			
	}
	
	if (!(procfile_page = (void *) get_zeroed_page(GFP_KERNEL)))
	{
		printk(LOG": Unable to allocate procfile page\n");
		err = -ENOMEM;
		goto out;
	}
	
	/* create the proc file */
	if (!(res->proc_entry = create_proc_entry(procfile, 0666, NULL)))
	{
		printk(LOG": Unable to create proc file %s\n", procfile);
		err = -EINVAL;
		goto out;
	}
	res->proc_entry->write_proc = proc_get_userland;
	res->proc_entry->read_proc = proc_put_userland;
	res->proc_entry->owner = THIS_MODULE;
	
	/* workqueue for bootstrap maelstrom resources, like raw sockets */
	if (!(bootstrap_wq = create_fifo_wqueue("mael-init-wq", CPU_MASK_ALL, NC)))
	{
		err = -ENOMEM;
		goto out;
	}
		
	/* zero them out */
	if (!(res->mael_stages = kzalloc(res->stages * 
		sizeof(struct mael_stage_resources), GFP_KERNEL)))
	{
		err = -ENOMEM;
		goto out;
	}

	for (i = 0; i < res->stages; i++)
	{
		char wqname[64];
		struct init_mael_stage_struct init_stage;
		struct mael_stage_resources *stage_res = &res->mael_stages[i];
		cpu_mask = CPU_MASK_ALL;
#ifdef BIND_WORKQUEUES_TO_CPUS
#ifdef BIND_WORKQUEUES_TO_CORES
		cpu_mask = cpumask_of_cpu(logical_cpu_map[i % num_present_cpus()]);
#else
		/* turns out CPU ids are already layed out for this case */
		cpu_mask = cpumask_of_cpu(i % num_present_cpus());
#endif /* BIND_WORKQUEUES_TO_CORES */
#endif /* BIND_WORKQUEUES_TO_CPUS */
		
		if ((err = init_bins(&stage_res->bins, r, c)) != 0)
		{
			printk(LOG": Unable to init bins, error %d\n", err);
			goto out;
		}
		
		if ((err = init_recovery_tables(&stage_res->recovery_tables)) != 0)
		{
			printk(LOG": Unable to init recovery tables, error %d\n", err);
			goto out;
		}
		
		INIT_LITE_WORK(&init_stage.work, init_mael_stage_resources);
		init_stage.ret = 0;
		init_stage.res = stage_res;
		init_stage.local_addr = &res->addr;
		fifo_queue_work(bootstrap_wq, &init_stage.work);
		flush_fifo_wqueue(bootstrap_wq);
		if (init_stage.ret != 0)
		{
			printk(LOG": Bootstrap init failed, error %d\n", err);
			err = init_stage.ret;
			goto out;
		}
		
		sprintf(wqname, "mael-stage-%d", i);
		if (!(stage_res->wq = create_fifo_wqueue(wqname, cpu_mask, NC)))
		{
			err = -ENOMEM;
			goto out;
		}
	}
	/* no need for the bootstrap workqueue */
	destroy_fifo_wqueue(bootstrap_wq);
	bootstrap_wq = NULL;

	/* start the UDP fec receiver thread, prepare everything ahead of time */
	init_completion(&res->fec_decoder_rcv_done);
	init_completion(&res->fec_decoder_rcv_started);
	
	if ((err = kernel_thread(fec_decoder_rcv_runner, res, CLONE_FS | 
		CLONE_FILES | SIGCHLD)) < 0) /* pid is returned in err */
	{
		printk(LOG": Unable to start fec decoder thread, error %d\n", err);
		goto out;
	}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
	res->fec_decoder_rcv = find_task_by_pid(err);
#else
	res->fec_decoder_rcv = find_task_by_pid_ns(err, &init_pid_ns);
#endif
	wait_for_completion(&res->fec_decoder_rcv_started);
	
#ifdef UDP_RCV_CPU_ID
	cpu_mask = cpumask_of_cpu(UDP_RCV_CPU_ID);
#else
	cpu_mask = CPU_MASK_ALL;
#endif /* BIND_UDP_RECEIVER_TO_CPUS */
	set_cpus_allowed(res->fec_decoder_rcv, cpu_mask);
	set_user_nice(res->fec_decoder_rcv, NC);
	
#ifdef USE_KMEM_CACHE_WORK
	if (!(mael_skb_work_cache = KMEM_CACHE(mael_skb_work, 
			SLAB_HWCACHE_ALIGN | SLAB_PANIC)))
	{
		printk(LOG": Unable to allocate kmem-cache for work_struct_lite\n");
		err = -ENOMEM;
		goto out;
	}
#endif /* USE_KMEM_CACHE_WORK */
	
	/* prepare and register the netfilter hook */
	res->nf_hook.hook = maelstrom_nf_hook;
	res->nf_hook.hooknum = hook_chain;
	res->nf_hook.pf = PF_INET;
	res->nf_hook.priority = hook_prio;
	err = nf_register_hook(&res->nf_hook);
	
	out:
	if (err != 0)
		free_mael_resources(res);
	else
		printk(LOG": Maelstrom module loaded\n");
	if (bootstrap_wq)
		destroy_fifo_wqueue(bootstrap_wq);
	return err;
}

static void reset_mael_reports(void) { memset(&mael_rep, 0, sizeof(mael_rep)); }

static inline int __init init_maelstrom(void)
{
	reset_mael_reports();
	return init_mael_resources(&mael_res, cores, slot_len, xor_r, xor_c);
}

static inline void __exit exit_maelstrom(void)
{
	reset_mael_reports();
	free_mael_resources(&mael_res);
	printk(LOG": Maelstrom module unloaded\n");
}

module_init(init_maelstrom);
module_exit(exit_maelstrom);

MODULE_DESCRIPTION("Maelstrom (NSDI08) PEP kernel module v2.0");
MODULE_AUTHOR("Tudor Marian");
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(a)
#endif
MODULE_LICENSE("Dual BSD/GPL");
