
/*
 * Copyright (C) 2008 Tudor Marian (tudorm@cs.cornell.edu)
 */

#include "mersene.h"

/* Period parameters */  
#define N MERSENE_TWISTER_N
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

//static unsigned long mt[N]; /* the array for the state vector  */
//static int mti = N+1; /* mti==N+1 means mt[N] is not initialized */

/* a cleaner interface to pseudorandom numbers.  should be moved to
 * common at some point. */
void pseudorand_init(struct mersene_twister *t, unsigned int seed)
{
	t->mti = N+1; /* before init_genrand, probably can be unset */
	init_genrand(t, seed);
}

/* return (min <= n < max) */
int pseudorand(struct mersene_twister *t, int min, int max)
{
	int diff = max - min, num;
	if (diff == 0)
	{
		return max;
	}
	else if (diff < 0)
	{
		int tmp;
		tmp = min;
		min = max;
		max = tmp;
		diff = max - min;
	}
	num = ((genrand_int32(t) % diff) + diff) % diff;
	return num + min;
}

/* initializes mt[N] with a seed */
void init_genrand(struct mersene_twister *t, unsigned long s)
{
	t->mt[0] = s & 0xffffffffUL;
	for (t->mti = 1; t->mti < N; t->mti++)
	{
		t->mt[t->mti] = (1812433253UL * (t->mt[t->mti - 1] ^ 
				(t->mt[t->mti - 1] >> 30)) + t->mti);
		/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
		/* In the previous versions, MSBs of the seed affect   */
		/* only MSBs of the array mt[].                        */
		/* 2002/01/09 modified by Makoto Matsumoto             */
		t->mt[t->mti] &= 0xffffffffUL;
		/* for >32 bit machines */
	}
}

/* initialize by an array with array-length */
/* init_key is the array for initializing keys */
/* key_length is its length */
/* slight change for C++, 2004/2/26 */
void init_by_array(struct mersene_twister *t, 
		unsigned long init_key[], int key_length)
{
	int i, j, k;
	init_genrand(t, 19650218UL);
	i = 1;
	j = 0;
	k = (N > key_length ? N : key_length);
	for (; k; k--)
	{
		t->mt[i] = (t->mt[i] ^ ((t->mt[i - 1] ^ 
				(t->mt[i - 1] >> 30)) * 1664525UL)) 
				+ init_key[j] + j; /* non linear */
		t->mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
		i++;
		j++;
		if (i >= N)
		{
			t->mt[0] = t->mt[N-1];
			i = 1;
		}
		if (j >= key_length)
			j = 0;
	}
	for (k = N-1; k; k--)
	{
		t->mt[i] = (t->mt[i] ^ ((t->mt[i - 1] ^ 
				(t->mt[i - 1] >> 30)) * 1566083941UL)) - i; /* non linear */
		t->mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
		i++;
		if (i >= N)
		{
			t->mt[0] = t->mt[N-1];
			i = 1;
		}
	}

	t->mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */
}

/* generates a random number on [0,0xffffffff]-interval */
unsigned long genrand_int32(struct mersene_twister *t)
{
	unsigned long y;
	static unsigned long mag01[2] = { 0x0UL, MATRIX_A};
	/* mag01[x] = x * MATRIX_A  for x=0,1 */

	if (t->mti >= N)
	{ /* generate N words at one time */
		int kk;

		if (t->mti == N + 1) /* if init_genrand() has not been called, */
			init_genrand(t, 5489UL); /* a default initial seed is used */

		for (kk = 0; kk < N-M; kk++)
		{
			y = (t->mt[kk] & UPPER_MASK) | (t->mt[kk + 1] & LOWER_MASK);
			t->mt[kk] = t->mt[kk + M] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		for (; kk < N-1; kk++)
		{
			y = (t->mt[kk] & UPPER_MASK) | (t->mt[kk + 1] & LOWER_MASK);
			t->mt[kk] = t->mt[kk + (M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		y = (t->mt[N-1] & UPPER_MASK) | (t->mt[0] & LOWER_MASK);
		t->mt[N-1] = t->mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

		t->mti = 0;
	}

	y = t->mt[t->mti++];

	/* Tempering */
	y ^= (y >> 11);
	y ^= (y << 7) & 0x9d2c5680UL;
	y ^= (y << 15) & 0xefc60000UL;
	y ^= (y >> 18);

	return y;
}

/* generates a random number on [0,0x7fffffff]-interval */
long genrand_int31(struct mersene_twister *t)
{
	return (long) (genrand_int32(t) >> 1);
}
