
/*
 * Copyright (C) 2008 Tudor Marian (tudorm@cs.cornell.edu)
 */

#include "wqueue.h"

/* to flush the workqueue in an elegant fashion */
struct wq_lite_cork
{
	struct work_struct_lite work;
	struct completion done;
};

static void wq_lite_cork_fn(struct work_struct_lite *work)
{
	struct wq_lite_cork *wq = container_of(work, struct wq_lite_cork, work);
	complete(&wq->done);
}

/* this is annoying, since __wake_up_locked is not exported as a symbol 
 * one has to reimplement / copy it here, and it's such a icache waste
 */
static inline void do_wake_up_common(wait_queue_head_t *q, unsigned int mode,
		int nr_exclusive, int sync, void *key)
{
	struct list_head *tmp, *next;

	list_for_each_safe(tmp, next, &q->task_list)
	{
		wait_queue_t *curr = list_entry(tmp, wait_queue_t, task_list);
		unsigned flags = curr->flags;

		if (curr->func(curr, mode, sync, key) &&
				(flags & WQ_FLAG_EXCLUSIVE) && !--nr_exclusive)
			break;
	}
}

static inline void __do_wake_up_locked(wait_queue_head_t *q)
{
	do_wake_up_common(q, (TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE), 
			1, 0, NULL);
}

void fifo_queue_work(struct fifo_wqueue *wq, work_struct_lite_t *work)
{
	//unsigned long flags;
	//spin_lock_irqsave(&wq->lock, flags);
	spin_lock_bh(&wq->lock);
	
	list_add_tail(&work->entry, &wq->worklist);
	wq->items++;
	if (wq->thread_sleeping)
		__do_wake_up_locked(&wq->sleeper_list);
	
	spin_unlock_bh(&wq->lock);
	//spin_unlock_irqrestore(&wq->lock, flags);
}

void fifo_queue_work_in_bh(struct fifo_wqueue *wq, work_struct_lite_t *work)
{
	spin_lock(&wq->lock);
	list_add_tail(&work->entry, &wq->worklist);
	wq->items++;
	if (wq->thread_sleeping)
		__do_wake_up_locked(&wq->sleeper_list);
	spin_unlock(&wq->lock);
}

/* enter without holding the spinlock */
static void drain_fifo_wqueue(struct fifo_wqueue *wq)
{
	LIST_HEAD(head);
	struct work_struct_lite *work, *tmp;

	//unsigned long flags;

	//spin_lock_irqsave(&wq->lock, flags);
	spin_lock_bh(&wq->lock);
	if (!list_empty(&wq->worklist))
	{
		list_replace_init(&wq->worklist, &head);
		wq->items = 0; /* keep this together with the list operations */
	}
	//spin_unlock_irqrestore(&wq->lock, flags);
	spin_unlock_bh(&wq->lock);
	
	list_for_each_entry_safe(work, tmp, &head, entry)
	{
		work_func_lite_t fn = work->func;
		list_del(&work->entry);
		fn(work); /* the function releases the work struct */
	}
}

static int runner_thread(void *param)
{
	//unsigned long flags;
	DEFINE_WAIT(wait);
	struct fifo_wqueue *wq = (struct fifo_wqueue *) param;

	set_user_nice(current, wq->nice);
	
	//spin_lock_irqsave(&wq->lock, flags);
	spin_lock_bh(&wq->lock);
	wq->thread_sleeping = 0;
	//spin_unlock_irqrestore(&wq->lock, flags);
	spin_unlock_bh(&wq->lock);
		
	while (!kthread_should_stop())
	{
		//spin_lock_irqsave(&wq->lock, flags);
		spin_lock_bh(&wq->lock);
		
		/* can be safely w/o the lock held, since pointer updates are atomic */
		if (list_empty(&wq->worklist))
		{
			__add_wait_queue(&wq->sleeper_list, &wait);
			wq->thread_sleeping = 1;
			
			//spin_unlock_irqrestore(&wq->lock, flags);
			spin_unlock_bh(&wq->lock);
			
			set_current_state(TASK_INTERRUPTIBLE);
			schedule(); /* block */
			
			//spin_lock_irqsave(&wq->lock, flags);
			spin_lock_bh(&wq->lock);
			
			wq->thread_sleeping = 0;
			__remove_wait_queue(&wq->sleeper_list, &wait);
			__set_current_state(TASK_RUNNING);
		}
		//spin_unlock_irqrestore(&wq->lock, flags);
		spin_unlock_bh(&wq->lock);
		
		drain_fifo_wqueue(wq);
	}

	return 0;
}

void flush_fifo_wqueue(struct fifo_wqueue *wq)
{
	struct wq_lite_cork cork;

	if (wq->thread == current)
		drain_fifo_wqueue(wq);
	else
	{
		INIT_LITE_WORK(&cork.work, wq_lite_cork_fn);
		init_completion(&cork.done);
		fifo_queue_work(wq, &cork.work);
		wait_for_completion(&cork.done);
	}
}

void destroy_fifo_wqueue(struct fifo_wqueue *wq)
{
	flush_fifo_wqueue(wq);
	if (wq->thread)
	{
		kthread_stop(wq->thread);
		wq->thread = NULL;
	}
	kfree(wq);
}

struct fifo_wqueue *create_fifo_wqueue(char *name, cpumask_t mask, int nice)
{
	struct fifo_wqueue *wq = NULL;
	struct task_struct *tsk = NULL;
	int err = 0;

	if (!(wq = kzalloc(sizeof(*wq), GFP_KERNEL)))
		return NULL;

	spin_lock_init(&wq->lock);
	INIT_LIST_HEAD(&wq->worklist);
	init_waitqueue_head(&wq->sleeper_list);
	wq->nice = nice;
	wq->thread_sleeping = 1;

	tsk = kthread_create(runner_thread, wq, "%s", name);
	if (IS_ERR(tsk))
	{
		err = -ENOMEM;
		goto out;
	}
	set_cpus_allowed(tsk, mask); /* tsk->cpus_allowed = mask; */
	wq->thread = tsk;
	wake_up_process(wq->thread);

	out: //
	if ((err != 0) && (wq))
	{
		destroy_fifo_wqueue(wq);
		wq = NULL;
	}
	return wq;
}

