/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 * 
 * Like a workqueue, but with a configurable CPU mask, and also more efficient
 * in waking up the worker task --- in particular we get a 200Mbps improvement
 * over the traditional in-kernel workqueue with the same functionality.
 * 
 * !! Important notice: For performance consideration, it was designed NOT to
 * work from interrupt handlers (hardirq), i.e. locking is performed with
 * spin_lock_bh and spin_lock instead of spin_lock_irqsave. 
 * You have been warned.
 * 
 */

#ifndef _WQUEUE_H_
#define _WQUEUE_H_

#include <linux/rbtree.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/kthread.h>

struct work_struct_lite;

typedef void (*work_func_lite_t)(struct work_struct_lite *work);

typedef struct work_struct_lite
{
	struct list_head entry;
	work_func_lite_t func;
}work_struct_lite_t;

struct fifo_wqueue
{
	/* protected by the lock, using it is quite an improvement (~200Mbps),
	 * by default, the kernel always issues an expensive __do_wake_up_locked */
	int thread_sleeping; 
	spinlock_t lock;
	struct list_head worklist;
	unsigned long items;
	wait_queue_head_t sleeper_list;
	struct task_struct *thread; /* the thread doing all the work */
	int nice;
} ____cacheline_aligned;

#define INIT_LITE_WORK(_work, _func)					\
	do {											\
			INIT_LIST_HEAD(&(_work)->entry);			\
			(_work)->func = (_func);				\
	} while (0)

/* returns 0 if @work was already on a queue, non-zero otherwise */
void fifo_queue_work(struct fifo_wqueue *wq, work_struct_lite_t *work);

/* like fifo_queue_work, but to be called in bottom halfs (e.g. softirq) */
void fifo_queue_work_in_bh(struct fifo_wqueue *wq, work_struct_lite_t *work);

/* can only be called in task context --- it waits for a completion */
void flush_fifo_wqueue(struct fifo_wqueue *wq);

/* can only be called in task context --- it calls flush_fifo_wqueue */
void destroy_fifo_wqueue(struct fifo_wqueue *wq);

struct fifo_wqueue *create_fifo_wqueue(char *name, cpumask_t mask, int nice);


#endif // _WQUEUE_H_

